"""
Model equations and constraints:
Burden sharing
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    Var,
    Param,
    GeneralConstraint,
    RegionalSoftEqualityConstraint,
    Any,
    quant,
    value,
)


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Emissions and temperature equations and constraints

    Necessary variables:

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.burden_sharing_regime = Param(within=Any)

    ## Burden sharing scheme:
    m.burden_sharing_common_level = Var(m.t, units=quant.unit("fraction_of_GDP"))

    constraints.extend(
        [
            # Total costs: abatement + damage costs should be equal among regions as % GDP
            RegionalSoftEqualityConstraint(
                lambda m, t, r: m.rel_abatement_costs[t, r]
                + m.damage_costs[t, r]
                + m.rel_financial_transfer[t, r],
                lambda m, t, r: m.burden_sharing_common_level[t],
                "burden_sharing_regime_total_costs",
                ignore_if=lambda m, t, r: value(m.burden_sharing_regime)
                != "equal_total_costs"
                or m.year(t) > 2100,
            ),
            # Abatement costs: abatement costs should be equal among regions as % GDP
            RegionalSoftEqualityConstraint(
                lambda m, t, r: m.rel_abatement_costs[t, r],
                lambda m, t, r: m.burden_sharing_common_level[t],
                "burden_sharing_regime_abatement_costs",
                ignore_if=lambda m, t, r: value(m.burden_sharing_regime)
                != "equal_abatement_costs"
                # or m.year(t) > 2125,
            ),
        ]
    )

    ## Per capita convergence:
    # m.regional_per_cap_emissions = Var(
    #     m.t, m.regions, units=quant.unit("emissionsrate_unit/population_unit")
    # )
    m.percapconv_share_init = Param(
        m.regions,
        initialize=lambda m, r: m.baseline_emissions(m.year(0), r)
        / sum(m.baseline_emissions(m.year(0), s) for s in m.regions),
    )
    m.percapconv_year = Param(initialize=2050)
    m.percapconv_share_pop = Param(
        m.t,
        m.regions,
        initialize=lambda m, t, r: m.population(m.year(t), r)
        / sum(m.population(m.year(t), s) for s in m.regions),
    )

    m.percapconv_emission_share = Var(
        m.t, m.regions  # , initialize=_percapconv_share_rule
    )
    m.percapconv_import_export_emission_reduction_balance = Var(
        m.t, m.regions, units=quant.unit("emissionsrate_unit")
    )
    constraints.extend(
        [
            RegionalSoftEqualityConstraint(
                lambda m, t, r: percapconv_share_rule(m, t, r) * m.global_emissions[t],
                lambda m, t, r: m.baseline[t, r] - m.paid_for_emission_reductions[t, r],
                epsilon=None,
                absolute_epsilon=0.01,
                ignore_if=lambda m, t, r: value(m.burden_sharing_regime)
                != "per_cap_convergence"
                or t == 0,
                name="percapconv_rule",
            ),
        ]
    )

    return constraints


def percapconv_share_rule(m, t, r):
    year_0, year_t = m.year(0), m.year(t)
    year_conv = m.percapconv_year

    if year_conv is False:
        # If it is false, use grandfathering all the time
        return m.percapconv_share_init[r]
    if year_conv == year_0:
        # If it is equal to first year, use immediate per capita convergence
        return m.percapconv_share_pop[t, r]

    year_linear_part = (year_t - year_0) / (year_conv - year_0)

    return (
        min(year_linear_part, 1) * m.percapconv_share_pop[t, r]
        + max(1 - year_linear_part, 0) * m.percapconv_share_init[r]
    )



"""
Model equations and constraints:
Financial transfer for damage costs module
Type: global damage cost pool
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    GeneralConstraint,
    RegionalConstraint,
    GlobalConstraint,
    RegionalInitConstraint,
    Constraint,
    Var,
)


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Damage cost trading equations and constraints
    (no-trade specification)

    Necessary variables:
        - m.financial_transfer

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    # Note: positive financial transfer means paying money, negative means receiving
    m.financial_transfer = Var(m.t, m.regions, initialize=0)
    m.rel_financial_transfer = Var(m.t, m.regions, initialize=0)

    constraints.extend(
        [
            GlobalConstraint(
                lambda m, t: sum(m.financial_transfer[t, r] for r in m.regions) == 0.0
                if t > 0
                else Constraint.Skip,
                "zero_sum_of_yearly_financial_transfer",
            ),
            RegionalInitConstraint(
                lambda m, r: m.financial_transfer[0, r] == 0.0,
                "no_transfer_in_first_year",
            ),
            RegionalConstraint(
                lambda m, t, r: m.financial_transfer[t, r]
                >= -m.damage_costs[t, r] * m.GDP_gross[t, r],
                "received_financial_transfer_max_own_damages",
            ),
            RegionalConstraint(
                lambda m, t, r: m.rel_financial_transfer[t, r] * m.GDP_gross[t, r]
                == m.financial_transfer[t, r],
                "financial_transfer_rel_to_gdp",
            ),
        ]
    )

    return constraints



"""
Imports all common modules and extra imports
"""

# Logging
import logging

logger = logging.getLogger("MIMOSA")

# General imports
from dataclasses import dataclass

# Pyomo imports
from pyomo.environ import (
    AbstractModel,
    ConcreteModel,
    Set,
    TransformationFactory,
    SolverFactory,
    SolverStatus,
    SolverManagerFactory,
    Objective,
    Param,
    Var,
    log,
    sqrt,
    tanh,
    value,
    maximize,
    minimize,
    NonNegativeReals,
    Any,
    units as u,
)
from pyomo.opt.base.solvers import OptSolver

# Pyomo utils
from .pyomo_utils import (
    Constraint,  # Only for Constraint.Skip
    GeneralConstraint,
    GlobalConstraint,
    GlobalInitConstraint,
    RegionalConstraint,
    RegionalInitConstraint,
    GlobalSoftEqualityConstraint,
    RegionalSoftEqualityConstraint,
    UsefulVar,
    soft_switch,
    soft_min,
    soft_max,
    get_all_variables,
    add_constraint,
    is_regional,
    get_indices,
    atan,
    exp,
)

# Other utils
from .utils import first

from .units import quant

# # Datastore
# from .data import DataStore

# # Config
# from .config import params



"""
Abstract representation of the model
--------------------------------------------
Contains all model equations and constraints
"""

import numpy as np
from mimosa.common import Param, AbstractModel, Set, add_constraint
from mimosa.components import (
    burdensharing,
    emissions,
    abatement,
    emissiontrade,
    financialtransfer,
    cobbdouglas,
    damages,
    objective,
    sealevelrise,
    welfare,
)


######################
# Create model
######################


def create_abstract_model(
    damage_module: str,
    emissiontrade_module: str,
    financialtransfer_module: str,
    welfare_module: str,
    objective_module: str,
) -> AbstractModel:
    m = AbstractModel()

    ## Constraints
    # Each constraint will be put in this list,
    # then added to the model at the end of this file.
    constraints = []

    ## Time and region
    m.beginyear = Param()
    m.dt = Param()
    m.tf = Param()
    m.t = Set()
    m.year = None  # Initialised with concrete instance
    m.year2100 = Param()

    m.regions = Set(ordered=True)

    ######################
    # Create data functions
    # Will be initialised when creating a concrete instance of the model
    ######################

    m.baseline_emissions = lambda year, region: None
    m.population = lambda year, region: None
    m.TFP = lambda year, region: None
    m.GDP = lambda year, region: None
    m.carbon_intensity = lambda year, region: None

    def baseline_cumulative(year_start, year_end, region):
        years = np.linspace(year_start, year_end, 100)
        return np.trapz(m.baseline_emissions(years, region), x=years)

    m.baseline_cumulative = baseline_cumulative
    m.baseline_cumulative_global = lambda m, year_start, year_end: sum(
        baseline_cumulative(year_start, year_end, r) for r in m.regions
    )

    ######################
    # Components
    ######################

    # Emissions and temperature equations
    constraints.extend(emissions.get_constraints(m))

    # Sea level rise
    constraints.extend(sealevelrise.get_constraints(m))

    # Damage costs
    if damage_module == "RICE2010":
        constraints.extend(damages.ad_rice2010.get_constraints(m))
    elif damage_module == "WITCH":
        constraints.extend(damages.ad_witch.get_constraints(m))
    elif damage_module == "RICE2012":
        constraints.extend(damages.ad_rice2012.get_constraints(m))
    elif damage_module == "COACCH":
        constraints.extend(damages.coacch.get_constraints(m))
    elif damage_module == "nodamage":
        constraints.extend(damages.nodamage.get_constraints(m))
    else:
        raise NotImplementedError

    # Abatement costs
    constraints.extend(abatement.get_constraints(m))

    # Emission trading
    if emissiontrade_module == "notrade":
        constraints.extend(emissiontrade.notrade.get_constraints(m))
    elif emissiontrade_module == "globalcostpool":
        constraints.extend(emissiontrade.globalcostpool.get_constraints(m))
    elif emissiontrade_module == "emissiontrade":
        constraints.extend(emissiontrade.emissiontrade.get_constraints(m))
    else:
        raise NotImplementedError(
            f"Emission trading module `{emissiontrade_module}` not implemented"
        )

    # Financial transfer
    if financialtransfer_module == "notransfer":
        constraints.extend(financialtransfer.notransfer.get_constraints(m))
    elif financialtransfer_module == "globaldamagepool":
        constraints.extend(financialtransfer.globaldamagepool.get_constraints(m))
    else:
        raise NotImplementedError(
            f"Financial transfer module `{financialtransfer_module}` not implemented"
        )

    # Burden sharing regime
    constraints.extend(burdensharing.get_constraints(m))

    # Cobb-Douglas and economics
    constraints.extend(cobbdouglas.get_constraints(m))

    # Utility and welfare
    if welfare_module == "inequal_aversion_elasmu":
        constraints.extend(welfare.inequal_aversion_elasmu.get_constraints(m))
    elif welfare_module == "inequal_aversion_zero":
        constraints.extend(welfare.inequal_aversion_zero.get_constraints(m))
    elif welfare_module == "inequal_aversion_general":
        constraints.extend(welfare.inequal_aversion_general.get_constraints(m))
    else:
        raise NotImplementedError(f"Welfare module `{welfare_module}` not implemented")

    # Objective of optimisation
    if objective_module == "utility":
        objective_rule, objective_constraints = objective.utility.get_constraints(m)
    elif objective_module == "globalcosts":
        objective_rule, objective_constraints = objective.globalcosts.get_constraints(m)
    else:
        raise NotImplementedError

    constraints.extend(objective_constraints)

    ######################
    # Add constraints to abstract model
    ######################

    for constraint in constraints:
        add_constraint(m, constraint.to_pyomo_constraint(m), constraint.name)

    m.obj = objective_rule

    return m



"""
Utils
"""


from glob import glob

import numpy as np
from mimosa.common.data.utils import extrapolate
import pandas as pd
from scipy.interpolate import InterpolatedUnivariateSpline
from mimosa.common import logger


def read_csv(path):
    if "*" in path:
        paths = glob(path)
        if len(paths) != 1:
            logger.warning(
                "Warning: matched {} files instead of 1 for file pattern {}".format(
                    len(paths), path
                )
            )
        path = glob(path)[0]
    return pd.read_csv(path)


class InterpolatingData:
    """
    `data` is a dataframe with as row-indices the regions (or "Global"), and as
    columns the years used in the ConcreteModel
    """

    def __init__(self, data):
        # extrapolate
        self.maxyear = float(data.columns[-1])
        self.region_interp_fct = {
            region: self.interp(row) for region, row in data.iterrows()
        }

    @staticmethod
    def interp(row):
        values = row.values
        years = row.index.astype(float).to_numpy()
        extra_years = np.arange(years[-1] + 5, 2151, 5)

        new_values = extrapolate(values, years, extra_years, "for_simulation_mode")
        all_years = np.concatenate([years, extra_years])
        return InterpolatedUnivariateSpline(all_years, new_values)

    def get(self, region, year):
        value = self.region_interp_fct[region](year)
        if len(value.shape) == 0:
            return float(value)
        return value



"""
Certain constraints can be deactivated. These are defined in:
    params["simulation"]["disabled_constraints"] = ['carbon_budget', ...]
"""


def deactivate_constraints(m, params):
    if params["simulation"]["deactivated_constraints"] is not None:
        for constraint_name in params["simulation"]["deactivated_constraints"]:
            constraint = getattr(m, f"constraint_{constraint_name}")
            if constraint is not None:
                constraint.deactivate()



import numpy as np
import pandas as pd
from collections import namedtuple

from mimosa.common import regional_params, data
from mimosa.concrete_model.instantiate_params import InstantiatedModel
from mimosa.concrete_model.simulation_mode.set_constraints import _get_interp_data
from mimosa.components import sealevelrise
from mimosa.components.damages.coacch import damage_fct


class SCC:
    # pylint: disable=no-member
    def __init__(self, params: dict):
        self.params = params
        self.regions = params["regions"]

        # Create the regional parameter store
        self.regional_param_store = regional_params.RegionalParamStore(self.params)

        # Create the data store
        self.data_store = data.DataStore(self.params, self.regional_param_store)

        self.data_functions = DataFunctions()

        instantiated_model = InstantiatedModel(
            self.data_functions,
            self.regional_param_store,
            self.data_store,
            False,
        )

        # From the instantiated model, only the parameter data ("instance_data")
        # is used. Note that we use the default namespace parameters, "None"
        instance_data = instantiated_model.instance_data[None]
        self.data = dict_to_object(instance_data)

        self.regions = self.data.regions
        self.years_i = list(self.data.t)
        self.years = [self.data_functions.year(i) for i in self.years_i]

        self.check_regional_emissions_input()

        self.calculate_damages()

    def check_regional_emissions_input(self):
        # Check if regional_emissions is given in params["simulation"]["constraint_variables"],
        # then use baseline regional emissions
        try:
            filepath_or_data = self.params["simulation"]["constraint_variables"][
                "regional_emissions"
            ]
            self.override_regional_emissions = True
            data_cache = {}
            interp_data = _get_interp_data(
                filepath_or_data, data_cache, "regional_emissions"
            )
            self.new_regional_emissions = pd.DataFrame(
                {r: interp_data.get(r, self.years) for r in self.regions},
                index=self.years,
            ).T
        except KeyError:
            self.override_regional_emissions = False

    def calculate_damages(self):
        out = {}

        years = self.years
        regions = self.regions

        # 1. Calculate regional emissions and GDP
        if self.override_regional_emissions:
            out["regional_emissions"] = self.new_regional_emissions

        else:
            out["regional_emissions"] = pd.DataFrame(
                {r: self.data_functions.baseline_emissions(years, r) for r in regions},
                index=years,
            ).T

        out["gdp"] = pd.DataFrame(
            {r: self.data_functions.GDP(years, r) for r in regions},
            index=years,
        ).T

        # 2. Calculate global emissions (sum of regional emissions) and cumulative emissions
        out["global_emissions"] = out["regional_emissions"].sum()
        cumulative_emissions = pd.Series(
            {
                year: np.trapz(
                    out["global_emissions"][:year],
                    x=out["global_emissions"][:year].index,
                )
                for year in out["global_emissions"].index
            }
        )

        # 3. From this, calculate temperature increase
        T0 = self.data.T0
        TCRE = self.data.TCRE
        out["temperature"] = T0 + TCRE * cumulative_emissions

        # 4. And calculate SLR
        total_slr = pd.Series()
        for t in self.years_i:
            if t == 0:
                slr_thermal = sealevelrise.slr_thermal_expansion_init(self.data)
                slr_cumgsic = 0.015
                slr_cumgis = 0.006
            else:
                prev_temp = out["temperature"][years[t - 1]]
                slr_thermal = sealevelrise.slr_thermal_expansion(
                    slr_thermal, prev_temp, self.data
                )
                slr_cumgsic = sealevelrise.slr_gsic(slr_cumgsic, prev_temp, self.data)
                slr_cumgis = sealevelrise.slr_gis(slr_cumgis, prev_temp, self.data)
            total_slr[years[t]] = slr_thermal + slr_cumgsic + slr_cumgis

        # 5. Calculate regional damages
        out["damages_temperature"] = pd.DataFrame(
            {
                r: self.data.damage_scale_factor
                * damage_fct(
                    out["temperature"] - 0.6,
                    out["temperature"].iloc[0] - 0.6,
                    self.data,
                    r,
                    False,
                )
                for r in regions
            }
        ).T

        out["damages_slr"] = pd.DataFrame(
            {
                r: self.data.damage_scale_factor
                * damage_fct(total_slr, total_slr.iloc[0], self.data, r, True)
                for r in regions
            }
        ).T

        out["damages_total"] = out["damages_temperature"] + out["damages_slr"]
        out["damages_absolute"] = out["damages_total"] * out["gdp"]

        self.output = out


class DataFunctions:
    pass


def dict_to_object(data_dict: dict, name="object"):
    obj = namedtuple(name, data_dict.keys())(
        *[v[None] if list(v.keys()) == [None] else v for v in data_dict.values()]
    )
    return obj



"""
Parsers for YAML config file:

- GeneralParser (*)
    - StringParser
    - FilepathParser
    - BoolParser
    - NumParser (*)
        - FloatParser
        - IntParser
    - EnumParser
    - QuantityParser
    - ListParser
        - DictParser

(*): Abstract class

These parsers are registered in the PARSER_FACTORY with their
associated parser type name (e.g. "str" for StringParser).
"""

import numpy as np
from abc import ABC, abstractmethod, abstractproperty
from pint import (
    DimensionalityError,
    OffsetUnitCalculusError,
    UndefinedUnitError,
    DefinitionSyntaxError,
)
from mimosa.common.units import Quantity
from . import get_nested


class GeneralParser(ABC):
    def __init__(self, node: dict, *args, **kwargs):
        try:
            self.type = node["type"]
        except KeyError:
            raise KeyError(f"Node should contain a type: {node}")

        self.descr = node.get("descr", "")
        self.default = node.get("default", None)
        self.can_be_false = node.get("can_be_false", False)

    def error(self, error_message):
        raise ValueError(
            f"""
            {error_message}
            Type: {self.type}
            Description: {self.descr}
            Default value: {self.default}
            Can be false: {self.can_be_false}
        """
        )

    @abstractmethod
    def parse(self, value):
        pass

    def get(self, user_config: dict, keys: list):
        try:
            to_be_parsed = get_nested(user_config, keys)
        except KeyError:
            # If it is not defined in the user config file, use the default value
            to_be_parsed = self.default
        return self.parse(to_be_parsed)

    def check_false(self, value):
        if self.can_be_false and self.is_falsy(value):
            return True
        return False

    def __repr__(self):
        return f"{self.__class__}: ({self.descr}). Default: {self.default}"

    @staticmethod
    def is_truthy(value):
        return value in [1, True, "True", "true", "yes"]

    @staticmethod
    def is_falsy(value):
        return value in [0, False, "False", "false", "no"]


class StringParser(GeneralParser):
    def parse(self, value):
        if self.check_false(value):
            return False
        return str(value)


class StringOrPlainDictParser(GeneralParser):
    def parse(self, value):
        if self.check_false(value):
            return False

        if isinstance(value, dict):
            return value

        return str(value)


class FilepathParser(GeneralParser):
    def parse(self, value):
        if self.check_false(value):
            return False
        # Should check whether file exists here
        return str(value)


class BoolParser(GeneralParser):
    def parse(self, value):
        if self.check_false(value):
            return False

        if self.is_truthy(value):
            parsed_value = True
        elif self.is_falsy(value):
            parsed_value = False
        else:
            self.error(f"Value {value} can not be parsed to a valid boolean")
        return parsed_value


class NumParser(GeneralParser):
    @abstractproperty
    def num_type_fct(self):
        pass

    def __init__(self, node, *args, **kwargs):
        GeneralParser.__init__(self, node, *args, **kwargs)
        self.min = node.get("min", -np.inf)
        self.max = node.get("max", np.inf)

    def parse(self, value):
        if self.check_false(value):
            return False
        try:
            parsed_value = self.num_type_fct(value)
        except ValueError:
            self.error(
                f"Value {value} can not be parsed to a {self.num_type_fct.__name__}"
            )
        # Check if value within bounds
        if parsed_value < self.min or parsed_value > self.max:
            self.error(
                f"Value {parsed_value} not within bounds ({self.min}, {self.max})"
            )
        return parsed_value


class FloatParser(NumParser):
    @property
    def num_type_fct(self):
        return float


class IntParser(NumParser):
    @property
    def num_type_fct(self):
        return int


class EnumParser(GeneralParser):
    def __init__(self, node, *args, **kwargs):
        GeneralParser.__init__(self, node, *args, **kwargs)
        self.allowed_values = node["values"]

    def parse(self, value):
        if self.check_false(value):
            return False
        if value not in self.allowed_values:
            self.error(f"Value {value} not in allowed values {self.allowed_values}")
        return value


class QuantityParser(GeneralParser):
    def __init__(self, node, quant: Quantity, *args, **kwargs):
        GeneralParser.__init__(self, node, *args, **kwargs)
        self.unit = node["unit"]
        self.quant = quant

    def parse(self, value):
        if self.check_false(value):
            return False
        # Try to parse the quantity
        try:
            parsed = self.quant(value, self.unit, can_be_false=False)
        except (
            DimensionalityError,
            DefinitionSyntaxError,
            OffsetUnitCalculusError,
            UndefinedUnitError,
        ):
            self.error(f"Cannot parse quantity `{value}` to unit `{self.unit}`")
        return value  # Returns the string, unit will be really converted when instantiating the model


class ListParser(GeneralParser):
    def __init__(self, node, quant, *args, **kwargs):
        GeneralParser.__init__(self, node, quant, *args, **kwargs)
        self.values_parser = (
            PARSER_FACTORY.create_parser(node["values"], quant)
            if "values" in node
            else None
        )

    @staticmethod
    def parse_if_not_none(parser: GeneralParser, value):
        return parser.parse(value) if parser is not None else value

    def parse(self, value):
        if self.check_false(value):
            return False
        if isinstance(value, list):
            parsed_list = [
                self.parse_if_not_none(self.values_parser, list_value)
                for list_value in value
            ]
            return parsed_list
        elif value is None:
            return []
        self.error(f"Cannot parse list {value}")


class DictParser(ListParser):
    def __init__(self, node, quant, *args, **kwargs):
        ListParser.__init__(self, node, quant, *args, **kwargs)
        # DictParser has an extra keys-parser compared to ListParser
        self.keys_parser = (
            PARSER_FACTORY.create_parser(node["keys"], quant)
            if "keys" in node
            else None
        )

    def parse(self, value):
        if self.check_false(value):
            return False
        if isinstance(value, dict):
            parsed_dict = {
                self.parse_if_not_none(
                    self.keys_parser, dict_key
                ): self.parse_if_not_none(self.values_parser, dict_value)
                for dict_key, dict_value in value.items()
            }
            return parsed_dict
        elif value is None:
            return {}
        self.error(f"Cannot parse dictionary {value}")


class ParserFactory:
    def __init__(self):
        self._parsers = {}

    def register_parser(self, typename: str, parser: GeneralParser):
        self._parsers[typename] = parser

    def create_parser(self, node: dict, quant: Quantity) -> GeneralParser:
        typename = node["type"]
        try:
            parser = self._parsers[typename]
        except KeyError:
            raise KeyError("Cannot find parser of type {}".format(typename))
        return parser(node, quant)


PARSER_FACTORY = ParserFactory()
PARSER_FACTORY.register_parser("str", StringParser)
PARSER_FACTORY.register_parser("str_or_plain_dict", StringOrPlainDictParser)
PARSER_FACTORY.register_parser("filepath", FilepathParser)
PARSER_FACTORY.register_parser("bool", BoolParser)
PARSER_FACTORY.register_parser("int", IntParser)
PARSER_FACTORY.register_parser("float", FloatParser)
PARSER_FACTORY.register_parser("enum", EnumParser)
PARSER_FACTORY.register_parser("quantity", QuantityParser)
PARSER_FACTORY.register_parser("list", ListParser)
PARSER_FACTORY.register_parser("dict", DictParser)



"""
Create a DataStore object which reads and parses regional data
from a data file in IIASA database format
"""

import os
from typing import Callable, Dict
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from mimosa.common import economics, quant
from .utils import UnitValues, extrapolate


class DataStore:
    """The DataStore object is used when instantiating a concrete model from the
    abstract mimosa. It provides values for time and region dependent data:
      - baseline emissions
      - population
      - baseline GDP / carbon intensity
      - TFP
    This data is read from a data file in IIASA format, and automatically transformed
    to the right units as specified in the config file.

    Usage: once the data store is initialised, only the method `DataStore.data_object` will
    be used.
    """

    # Class property and not instance property to reduce redundancy
    databases = {}
    cached_data = {}

    def __init__(self, params, regional_param_store):
        self.params = params
        self._select_database()
        self._create_data_years()

        baseline = self._create_data_values("emissions", "emissionsrate_unit")
        population = self._create_data_values("population", "population_unit")
        gdp = self._create_data_values("GDP", "currency_unit")
        self._data_values = {
            "baseline": baseline,
            "population": population,
            "GDP": gdp,
            "carbon_intensity": {
                r: UnitValues(baseline[r].xvalues, baseline[r].yvalues / gdp[r].yvalues)
                for r in params["regions"]
            },
            "TFP": {
                r: economics.get_TFP(
                    r, self.data_years, gdp, population, regional_param_store
                )
                for r in params["regions"]
            },
        }

    def _select_database(self):
        """Makes sure the file doesn't need to be read multiple times"""
        filename = os.path.join(
            os.path.dirname(__file__),
            "../../",
            self.params["input"]["db_filename"],
        )
        if filename not in DataStore.databases:
            database = pd.read_csv(filename)
            database.columns = database.columns.str.lower()
            DataStore.databases[filename] = database
            self.cached_data[filename] = {}

        self.database = DataStore.databases[filename]
        self.cache = self.cached_data[filename]
        self.filename = filename

    def _create_data_years(self):
        beginyear = self.params["time"]["start"]
        endyear = self.params["time"]["end"] + 50  # Make sure extrapolation goes well
        dt = self.params["time"]["dt"]
        self.data_years = np.arange(beginyear, endyear, dt)

    def _create_data_values(self, variable, to_unit=None) -> Dict[str, UnitValues]:
        regions = self.params["regions"]
        return {
            region: self._get_data(self.data_years, region, variable, to_unit)
            for region in regions
        }

    def _get_data_from_database(self, region, variable):
        SSP = self.params["SSP"]
        model = self.params["input"]["baselines"][SSP]["model"]
        scenario = self.params["input"]["baselines"][SSP]["scenario"]
        variablename = self.params["input"]["variables"][variable]

        cached_data = self._get_cached_data_from_database(
            model, scenario, region, variablename
        )

        return cached_data["years"], cached_data["values"], cached_data["unit"]

    def _get_cached_data_from_database(self, model, scenario, region, variablename):
        key = (model, scenario, region, variablename)
        if key not in self.cache:
            database = self.database
            selection = database.loc[
                (database["model"] == model)
                & (database["scenario"] == scenario)
                & (database["region"] == (region[:-1] if region[-1] == "#" else region))
                & (database["variable"] == variablename)
            ]

            if len(selection) != 1:
                raise Exception(
                    "Found {} rows matching criteria ({}, {}, {}, {}) instead of one.".format(
                        len(selection), model, scenario, region, variablename
                    )
                )

            self.cache[key] = {
                "years": selection.loc[:, "2010":].columns.values.astype(float),
                "values": selection.loc[:, "2010":].values[0],
                "unit": selection.iloc[0]["unit"],
            }
        return self.cache[key]

    def _get_data(self, output_years, region, variable, to_unit=None) -> UnitValues:
        # 1. Get data from database
        years, values, unit = self._get_data_from_database(region, variable)

        if to_unit is not None:
            quantity = quant(values, unit, to_unit, only_magnitude=False)
            values = quantity.magnitude
            unit = quantity.units

        # 2. Extrapolate this data to beyond 2100
        extra_years = np.arange(2110, 2260, 10)
        extended_data = extrapolate(values, years, extra_years, [variable, region])
        extended_years = np.concatenate([years, extra_years])

        # 3. Interpolate the combined data
        interp_fct = interp1d(extended_years, extended_data, kind="cubic")
        return UnitValues(output_years, interp_fct(output_years), unit)

    def _interp_data(self, year, region, variable):
        values = self._data_values[variable][region]
        return np.interp(year, values.xvalues, values.yvalues)

    def data_object(self, variable: str) -> Callable[[int, str], float]:
        """Creates a function giving the value of `variable` at a given year and regions.

        Args:
            variable (str): any of the keys of self._data_values

        Returns:
            Callable[[int, str], float]: interpolating function of type f(year, region)
        """
        return lambda year, region: self._interp_data(year, region, variable)

    def __repr__(self):
        return "DataStore with data values {} calculated on the years {}-{} from input file {} for the regions {} and {}".format(
            list(self._data_values.keys()),
            self.data_years[0],
            self.data_years[-1],
            self.filename,
            list(self.params["regions"].keys()),
            self.params["SSP"],
        )



"""
Creates the class MIMOSA:
This is the main class. It builds a new AbstractModel
using the chosen damage and objective modules, then reads in the
parameter values and data (from the DataStore). With these values,
it creates an `instance` of the AbstractModel. This is then sent to the solver.
Finally, the export functions are called here.
"""

import os

from mimosa.common import (
    AbstractModel,
    ConcreteModel,
    TransformationFactory,
    SolverFactory,
    SolverManagerFactory,
    SolverStatus,
    value,
    OptSolver,
    data,
    regional_params,
    utils,
    units,
    logger,
)
from mimosa.common.config.parseconfig import check_params
from mimosa.export import full_plot, visualise_ipopt_output, save_output
from mimosa.abstract_model import create_abstract_model
from mimosa.concrete_model.instantiate_params import InstantiatedModel
from mimosa.concrete_model import simulation_mode


class MIMOSA:
    """
    The MIMOSA object creates an AbstractModel, makes a Concrete instance
    out of it, solves it and saves the results.

    Args:
        params (dict): contains all Param values, and is based on `input/config.yaml`

    Attributes:
        params (dict)
        regions (dict): taken from params
        abstract_model (AbstractModel): the AbstractModel created using the chosen damage/objective modules
        data_store (DataStore): object used to access regional data from the input database
        m (ConcreteModel): concrete instance of `abstract_model`

    """

    def __init__(self, params: dict):
        # Check if input parameter dictionary is valid
        params = check_params(params)
        self.params = params
        self.regions = params["regions"]

        self.abstract_model = self.get_abstract_model()
        self.concrete_model = self.create_instance()
        self.preprocessing()

    def get_abstract_model(self) -> AbstractModel:
        """
        Returns:
            AbstractModel: model corresponding to the damage/objective module combination
        """
        damage_module = self.params["model"]["damage module"]
        emissiontrade_module = self.params["model"]["emissiontrade module"]
        financialtransfer_module = self.params["model"]["financialtransfer module"]
        welfare_module = self.params["model"]["welfare module"]
        objective_module = self.params["model"]["objective module"]

        return create_abstract_model(
            damage_module,
            emissiontrade_module,
            financialtransfer_module,
            welfare_module,
            objective_module,
        )

    @utils.timer("Concrete model creation")
    def create_instance(self) -> ConcreteModel:
        """
        Creates the objects necessary for the concrete model:
          - Regional parameter store
          - Data store
        Using these, it transforms the AbstractModel into a ConcreteModel

        Returns:
            ConcreteModel: model instantiated with parameter values and data functions
        """

        # Create the regional parameter store
        self.regional_param_store = regional_params.RegionalParamStore(self.params)

        # Create the data store
        self.data_store = data.DataStore(self.params, self.regional_param_store)

        # Using these help objects, create the instantiated model
        instantiated_model = InstantiatedModel(
            self.abstract_model, self.regional_param_store, self.data_store
        )
        m = instantiated_model.concrete_model

        # When using simulation mode, add extra constraints to variables and disable other constraints
        if (
            self.params.get("simulation") is not None
            and self.params["simulation"]["simulationmode"]
        ):
            simulation_mode.set_simulation_mode(m, self.params)

        return m

    def preprocessing(self) -> None:
        """
        Pyomo can apply certain pre-processing steps before sending the model
        to the solver. These include:
          - Aggregate variables that are linked by equality constraints
          - Initialise non-fixed variables to midpoint of their boundaries
          - Fix variables that are de-facto fixed
          - Propagate variable fixing for equalities of type x = y
        """

        if len(self.regions) > 1:
            TransformationFactory("contrib.aggregate_vars").apply_to(
                self.concrete_model
            )
        TransformationFactory("contrib.init_vars_midpoint").apply_to(
            self.concrete_model
        )
        TransformationFactory("contrib.detect_fixed_vars").apply_to(self.concrete_model)
        if len(self.regions) > 1:
            TransformationFactory("contrib.propagate_fixed_vars").apply_to(
                self.concrete_model
            )

    def postprocessing(self) -> None:
        """Post-processing tasks to restore aggregate variables in pre-processing step"""
        if len(self.regions) > 1:
            TransformationFactory("contrib.aggregate_vars").update_variables(
                self.concrete_model
            )

    @utils.timer("Model solve", True)
    def solve(
        self,
        verbose=True,
        halt_on_ampl_error="no",
        use_neos=False,
        neos_email=None,
        ipopt_output_file=None,
    ) -> None:
        """Sends the concrete model to a solver.

        Args:
            verbose (bool, optional): Prints intermediate IPOPT results. Defaults to True.
            halt_on_ampl_error (str, optional): Lets IPOPT stop when invalid values are encountered. Defaults to "no".
            use_neos (bool, optional): Uses the external NEOS server for solving. Defaults to False.
            neos_email (str or None, optional): E-mail address for NEOS server. Defaults to None.
            ipopt_output_file (str or None, optional): Filename for IPOPT intermediate output. Defaults to None.

        Raises:
            SolverException: raised if solver did not exit with status OK
        """

        if use_neos:
            # Send concrete model to external solver on NEOS server
            # Requires authenticated email address
            os.environ["NEOS_EMAIL"] = neos_email
            solver_manager = SolverManagerFactory("neos")
            solver = "ipopt"  # or "conopt'
            results = solver_manager.solve(self.concrete_model, opt=solver)
        else:
            # Solve locally using ipopt
            opt: OptSolver = SolverFactory("ipopt")
            opt.options["halt_on_ampl_error"] = halt_on_ampl_error
            # opt.options["max_iter"] = 5
            if ipopt_output_file is not None:
                opt.options["output_file"] = ipopt_output_file
            results = opt.solve(
                self.concrete_model, tee=verbose, symbolic_solver_labels=True
            )
            if ipopt_output_file is not None:
                visualise_ipopt_output(ipopt_output_file)

        self.postprocessing()

        logger.info("Status: {}".format(results.solver.status))

        if results.solver.status != SolverStatus.ok:
            logger.error(
                "Status: {}, termination condition: {}".format(
                    results.solver.status, results.solver.termination_condition
                )
            )
            if results.solver.status != SolverStatus.warning:
                raise SolverException("Solver did not exit with status OK")

        logger.info(
            "Final NPV: {}".format(
                value(self.concrete_model.NPV[self.concrete_model.tf])
            )
        )

    @utils.timer("Plotting results")
    def plot(self, filename="result", **kwargs):
        full_plot(self.concrete_model, filename, **kwargs)

    def save(self, experiment=None, **kwargs):
        save_output(self.params, self.concrete_model, experiment, **kwargs)


###########################
##
## Utils
##
###########################


class SolverException(Exception):
    """Raised when Pyomo solver does not exit with status OK"""



"""
Model equations and constraints:
Emission trading module
Type: no trade
"""

from typing import Sequence
from mimosa.common import AbstractModel, GeneralConstraint, RegionalConstraint, Param

from mimosa.components.abatement import AC


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Emission trading equations and constraints
    (no-trade specification)

    Necessary variables:
        m.abatement_costs (abatement costs as paid for by this region)

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.import_export_emission_reduction_balance = Param(m.t, m.regions, initialize=0)

    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: (m.abatement_costs[t, r])
                == AC(m.relative_abatement[t, r], m, t, r) * m.baseline[t, r],
                "abatement_costs",
            ),
        ]
    )

    return constraints



"""The store uses RegionalParameters objects, defined here"""

import os
import pandas as pd

from .region_mappers import REGION_MAPPERS


class RegionalParamContainer:
    def __init__(self, filename, regiontype_input, regiontype_output):
        full_filename = os.path.join(
            os.path.dirname(__file__),
            "../../inputdata/params",
            filename,
        )
        data_raw = pd.read_csv(full_filename)

        if regiontype_input == regiontype_output:
            self.data = data_raw.set_index("region")
        else:
            # Transform the input regional definition to required output regional definitions
            try:
                region_mapper = REGION_MAPPERS[(regiontype_input, regiontype_output)]
            except KeyError:
                raise KeyError(
                    f"Region mapper missing between {regiontype_input} and {regiontype_output}"
                )

            self.data = region_mapper.map_regions(
                data_raw, regiontype_input, regiontype_output
            ).set_index("region")

    def get(self, paramname, region):
        value = self.data.loc[region, paramname]
        try:
            return float(value)
        except ValueError:
            # If unable to convert to float:
            return value



"""
Parses the config.yaml file and checks for consistency with the default config template.
"""


from mimosa.common.utils import load_yaml
from mimosa.common import quant

from .utils import PARSER_FACTORY, set_nested, get_nested, flatten


def parse_params(default_yaml, user_yaml, return_parser_tree=False):
    parsed_dict = {}
    parser_tree = {}

    def _recursive_traverse(curr_keys, subset):
        for key, node in subset.items():
            keys = list(curr_keys) + [key]
            if isinstance(node, dict) and "type" not in node:
                _recursive_traverse(keys, node)
            else:
                parser = PARSER_FACTORY.create_parser(node, quant)

                # Save the parser instance for possible later use
                set_nested(parser_tree, keys, parser)

                # Save the parsed value
                set_nested(parsed_dict, keys, parser.get(user_yaml, keys))

    _recursive_traverse([], default_yaml)

    if return_parser_tree:
        return parsed_dict, parser_tree

    return parsed_dict


def check_obsolete_params(user_yaml, parsed_params, parser_tree):
    def leaf_criterium(keys, node):
        try:
            parser_type = get_nested(parser_tree, keys).type
        except (AttributeError, KeyError):
            parser_type = None
        if isinstance(node, dict) and parser_type != "dict":
            return False
        return True

    keys_user = set(flatten(user_yaml, leaf_criterium=leaf_criterium))
    keys_parsed = set(flatten(parsed_params, leaf_criterium=leaf_criterium))

    # Check keys which are in user config, but not in parsed file
    # These are obsolete/unused/misspelled parameters
    num_obsolete = 0
    for key in set(keys_user) - set(keys_parsed):
        num_obsolete += 1
        print(f"Obsolete key: {key}")
    if num_obsolete > 0:
        raise RuntimeWarning("Some config parameters are obsolete.")


def check_params(input_params):
    default_yaml = load_yaml("config_default.yaml")

    parsed_params, params_parser_tree = parse_params(
        default_yaml, input_params, return_parser_tree=True
    )

    check_obsolete_params(input_params, parsed_params, params_parser_tree)

    return parsed_params


def load_params(user_yaml_filename="config.yaml"):
    user_yaml = load_yaml(user_yaml_filename)
    return check_params(user_yaml)



from dataclasses import dataclass
from typing import Iterable

import numpy as np


@dataclass
class UnitValues:
    xvalues: Iterable
    yvalues: Iterable
    unit: str = None


def extrapolate(input_values, years, extra_years, meta_info, stabilising_years=50):
    """
    To extrapolate: take growth rate 2090-2100, linearly bring it down to growth rate of 0 in 2150
    """

    became_negative = False

    # First get final change rate
    change_rate = (input_values[-1] - input_values[-2]) / (years[-1] - years[-2])
    minmax = np.maximum if change_rate > 0 else np.minimum

    t_prev = years[-1]
    val_prev = input_values[-1]

    new_values = []

    for t in extra_years:
        change = minmax(
            0.0, change_rate - change_rate * (t_prev - 2100.0) / stabilising_years
        )
        val = val_prev + change * (t - t_prev)

        # try:
        #     if meta_info[0] == "emissions":
        #         val = val_prev * (
        #             0.98 ** (t - t_prev)
        #         )  # Reduce 2% per year for emissions after 2100
        # except IndexError:
        #     pass
        # if val < 0:
        #     val = 0.1
        #     became_negative = True

        new_values.append(val)

        val_prev = val
        t_prev = t

    if became_negative:
        print("Extrapolation became negative for", meta_info)
    return np.concatenate([input_values, new_values])



"""
Model equations and constraints:
Emission trading module
Type: global cost pool
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    Var,
    Param,
    GeneralConstraint,
    GlobalConstraint,
    RegionalConstraint,
    Constraint,
    NonNegativeReals,
    quant,
)

from mimosa.components.abatement import AC


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Emission trading equations and constraints
    (global cost pool specification)

    Necessary variables:
        m.abatement_costs (abatement costs as paid for by this region)

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.area_under_MAC = Var(
        m.t,
        m.regions,
        within=NonNegativeReals,
        initialize=0,
        units=quant.unit("currency_unit"),
    )

    # The global mitigation cost pool:
    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.area_under_MAC[t, r]
                == AC(m.relative_abatement[t, r], m, t, r) * m.baseline[t, r],
                "abatement_costs",
            ),
            GlobalConstraint(
                lambda m, t: sum(m.abatement_costs[t, r] for r in m.regions)
                == sum(m.area_under_MAC[t, r] for r in m.regions),
                "sum_abatement_equals_sum_area_under_mac",
            ),
        ]
    )

    ## Extra reporting variables:

    m.paid_for_emission_reductions = Var(
        m.t, m.regions, units=quant.unit("emissionsrate_unit")
    )
    m.import_export_emission_reduction_balance = Var(
        m.t, m.regions, units=quant.unit("emissionsrate_unit")
    )
    m.import_export_mitigation_cost_balance = Var(
        m.t, m.regions, units=quant.unit("currency_unit")
    )
    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.paid_for_emission_reductions[t, r]
                == m.abatement_costs[t, r]
                * m.global_emission_reduction_per_cost_unit[t]
                if t > 0
                else Constraint.Skip,
                "paid_for_emission_reductions",
            ),
            # Import export of emission reduction balance: if positive: , if negative:
            RegionalConstraint(
                lambda m, t, r: m.import_export_emission_reduction_balance[t, r]
                == m.paid_for_emission_reductions[t, r]
                - m.regional_emission_reduction[t, r]
                if t > 0
                else Constraint.Skip,
                "import_export_emission_reduction_balance",
            ),
            RegionalConstraint(
                lambda m, t, r: m.import_export_mitigation_cost_balance[t, r]
                == m.abatement_costs[t, r] - m.area_under_MAC[t, r],
                "import_export_mitigation_cost_balance",
            ),
        ]
    )

    # How are mitigation costs distributed over regions?
    m.min_rel_payment_level = Param()
    m.max_rel_payment_level = Param()

    return constraints



"""
Model equations and constraints:
Economics and Cobb-Douglas
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    Param,
    Var,
    GeneralConstraint,
    RegionalConstraint,
    RegionalInitConstraint,
    Constraint,
    value,
    soft_min,
    economics,
    quant,
)


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Economics and Cobb-Douglas equations and constraints

    Necessary variables:
        m.L (equal to m.population)
        m.dk

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.init_capitalstock_factor = Param(m.regions, units=quant.unit("currency_unit"))
    m.capital_stock = Var(
        m.t,
        m.regions,
        initialize=lambda m, t, r: m.init_capitalstock_factor[r] * m.GDP(m.year(t), r),
        units=quant.unit("currency_unit"),
    )

    # Parameters
    m.alpha = Param()
    m.dk = Param()
    m.sr = Param()

    m.GDP_gross = Var(
        m.t,
        m.regions,
        initialize=lambda m, t, r: m.GDP(m.year(0), r),
        units=quant.unit("currency_unit"),
    )
    m.GDP_net = Var(
        m.t,
        m.regions,
        units=quant.unit("currency_unit"),
        initialize=lambda m, t, r: m.GDP(m.year(0), r),
    )
    m.investments = Var(m.t, m.regions, units=quant.unit("currency_unit"))
    m.consumption = Var(m.t, m.regions, units=quant.unit("currency_unit"))

    m.ignore_damages = Param()

    # Cobb-Douglas, GDP, investments, capital and consumption
    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.GDP_gross[t, r]
                == economics.calc_GDP(
                    m.TFP(m.year(t), r),
                    m.L(m.year(t), r),
                    soft_min(m.capital_stock[t, r], scale=10),
                    m.alpha,
                )
                if t > 0
                else Constraint.Skip,
                "GDP_gross",
            ),
            RegionalInitConstraint(
                lambda m, r: m.GDP_gross[0, r] == m.GDP(m.year(0), r), "GDP_gross_init"
            ),
            RegionalConstraint(
                lambda m, t, r: m.GDP_net[t, r]
                == m.GDP_gross[t, r]
                * (1 - (m.damage_costs[t, r] if not value(m.ignore_damages) else 0))
                - m.abatement_costs[t, r]
                - m.financial_transfer[t, r],
                "GDP_net",
            ),
            RegionalConstraint(
                lambda m, t, r: m.investments[t, r] == m.sr * m.GDP_net[t, r],
                "investments",
            ),
            RegionalConstraint(
                lambda m, t, r: m.consumption[t, r] == (1 - m.sr) * m.GDP_net[t, r],
                "consumption",
            ),
            RegionalConstraint(
                lambda m, t, r: (
                    m.capital_stock[t, r]
                    == m.capital_stock[t - 1, r]
                    + m.dt
                    * economics.calc_dKdt(
                        m.capital_stock[t, r], m.dk, m.investments[t, r], m.dt
                    )
                )
                if t > 0
                else Constraint.Skip,
                "capital_stock",
            ),
            RegionalInitConstraint(
                lambda m, r: m.capital_stock[0, r]
                == m.init_capitalstock_factor[r] * m.GDP(m.year(0), r)
            ),
        ]
    )

    return constraints



"""
Model equations and constraints:
Damage and adaptation costs, RICE specification
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    Param,
    Var,
    GeneralConstraint,
    GlobalConstraint,
    RegionalConstraint,
    RegionalInitConstraint,
    soft_min,
    value,
    tanh,
)


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Damage and adaptation costs equations and constraints
    (RICE specification)

    Necessary variables:
        m.damage_costs (sum of residual damages and adaptation costs, as % of GDP)

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.damage_costs = Var(m.t, m.regions)
    m.smoothed_factor = Var(m.t, bounds=(0, 1))
    m.gross_damages = Var(m.t, m.regions)
    m.resid_damages = Var(m.t, m.regions)
    m.adapt_costs = Var(m.t, m.regions)
    m.adapt_level = Var(m.t, m.regions, bounds=(0, 1))

    m.damage_a1 = Param(m.regions)
    m.damage_a2 = Param(m.regions)
    m.damage_a3 = Param(m.regions)
    m.damage_scale_factor = Param()
    m.adapt_g1 = Param(m.regions)
    m.adapt_g2 = Param(m.regions)
    m.adapt_curr_level = Param()
    m.fixed_adaptation = Param()

    constraints.extend(
        [
            # Smoothing factor for partially reversible damages
            GlobalConstraint(
                lambda m, t: m.smoothed_factor[t]
                == smoothed_factor(t, m.temperature, m.dt, m.perc_reversible_damages),
                name="smoothed_factor",
            ),
            # Gross damages
            RegionalConstraint(
                lambda m, t, r: m.gross_damages[t, r]
                == calc_gross_damages(m, t, r, m.perc_reversible_damages),
                name="gross_damages",
            ),
            RegionalConstraint(lambda m, t, r: m.gross_damages[t, r] >= 0),
            RegionalInitConstraint(lambda m, r: m.gross_damages[0, r] == 0),
        ]
    )

    # Adaptation
    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.adapt_level[t, r]
                == (
                    m.adapt_curr_level
                    if value(m.fixed_adaptation)
                    else optimal_adapt_level(m.gross_damages[t, r], m, r)
                ),
                name="adapt_level",
            ),
            RegionalConstraint(
                lambda m, t, r: m.resid_damages[t, r]
                == m.gross_damages[t, r] * (1 - m.adapt_level[t, r]),
                "resid_damages",
            ),
            RegionalConstraint(
                lambda m, t, r: m.adapt_costs[t, r]
                == adaptation_costs(m.adapt_level[t, r], m, r),
                "adapt_costs",
            ),
            RegionalConstraint(
                lambda m, t, r: m.damage_costs[t, r]
                == m.resid_damages[t, r] + m.adapt_costs[t, r],
                "damage_costs",
            ),
        ]
    )

    return constraints


#################
## Utils
#################


# Damage function


def calc_gross_damages(m, t, r, perc_reversible_damages):
    # If using partially irreversible damage, the damages
    # are calculated using a differential equation (delta_damage)
    if perc_reversible_damages < 1 and t > 0:
        delta_temperature = m.temperature[t] - m.temperature[t - 1]
        delta_damage = (
            damage_fct_dot(m.temperature[t], m, r)
            * m.smoothed_factor
            * delta_temperature
            / m.dt
        )
        return m.gross_damages[t - 1, r] + m.dt * m.damage_scale_factor * delta_damage

    # Otherwise, simply use the damage function
    return m.damage_scale_factor * damage_fct(m.temperature[t], m.T0, m, r)


def damage_fct(temperature, init_temp, m, r):
    power_fct = (
        lambda temp: m.damage_a1[r] * temp + m.damage_a2[r] * temp ** m.damage_a3[r]
    )

    damage = power_fct(soft_min(temperature))
    if init_temp is not None:
        damage -= power_fct(init_temp)

    return damage


def damage_fct_dot(temperature, m, r):
    """
    Derivative of the power function
    """
    return m.damage_a1[r] + m.damage_a2[r] * m.damage_a3[r] * temperature ** (
        m.damage_a3[r] - 1
    )


# Adaptation cost function
def adaptation_costs(adapt_level, m, r):
    return m.adapt_g1[r] * soft_min(adapt_level) ** m.adapt_g2[r]


def optimal_adapt_level(gross_damages, m, r):
    if value(m.adapt_g1[r]) * value(m.adapt_g2[r]) == 0:
        return 0

    eps = 0.0005
    return (soft_min(gross_damages, 0.01) / (m.adapt_g1[r] * m.adapt_g2[r]) + eps) ** (
        1 / (m.adapt_g2[r] - 1)
    )

    # return pow(GD / (m.adapt_g1[r] * m.adapt_g2[r]) + eps, 1/(m.adapt_g2[r]-1), abs=True)


# Partially irreversible damages:
def smoothed_factor(t, temperature, dt, perc_revers):
    if perc_revers < 1 and t > 0:
        delta_temp = temperature[t] - temperature[t - 1]
        return (tanh((delta_temp / dt) / 1e-3) + 1) * (
            1 - perc_revers
        ) / 2 + perc_revers

    return 1



"""
Model equations and constraints:
Damage and adaptation costs, RICE specification
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    Param,
    Var,
    GeneralConstraint,
    RegionalConstraint,
    value,
    soft_max,
    Any,
    exp,
    quant,
)


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Damage and adaptation costs equations and constraints
    (COACCH specification)

    Necessary variables:
        m.damage_costs (sum of residual damages and adaptation costs multiplied by gross GDP)

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.damage_costs = Var(m.t, m.regions, units=quant.unit("fraction_of_GDP"))
    m.damage_scale_factor = Param()

    # Damages not related to SLR (dependent on temperature)
    m.resid_damages = Var(m.t, m.regions, units=quant.unit("fraction_of_GDP"))

    m.damage_noslr_form = Param(m.regions, within=Any)  # String for functional form
    m.damage_noslr_b1 = Param(m.regions)
    m.damage_noslr_b2 = Param(m.regions)
    m.damage_noslr_b3 = Param(m.regions, within=Any)  # Can be empty
    # (b2 and b3 are only used for some functional forms)

    m.damage_noslr_a = Param(m.regions)

    # Quadratic damage function for non-SLR damages. Factor `a` represents
    # the damage quantile
    constraints.append(
        RegionalConstraint(
            lambda m, t, r: m.resid_damages[t, r]
            == m.damage_scale_factor
            * damage_fct(m.temperature[t] - 0.6, m.T0 - 0.6, m, r, is_slr=False),
            "resid_damages",
        )
    )

    # SLR damages
    m.SLR_damages = Var(
        m.t, m.regions, bounds=(-0.5, 0.7), units=quant.unit("fraction_of_GDP")
    )

    m.damage_slr_form = Param(m.regions, within=Any)  # String for functional form
    m.damage_slr_b1 = Param(m.regions)
    m.damage_slr_b2 = Param(
        m.regions, within=Any
    )  # within=Any since it can be empty for some functional forms
    m.damage_slr_b3 = Param(
        m.regions, within=Any
    )  # within=Any since it can be empty for some functional forms
    # (b2 and b3 are only used for some functional forms)

    m.damage_slr_a = Param(m.regions)

    # Linear damage function for SLR damages, including adaptation costs
    constraints.append(
        RegionalConstraint(
            lambda m, t, r: m.SLR_damages[t, r]
            == m.damage_scale_factor
            * damage_fct(m.total_SLR[t], m.total_SLR[0], m, r, is_slr=True),
            "SLR_damages",
        )
    )

    # Total damages are sum of non-SLR and SLR damages
    constraints.append(
        RegionalConstraint(
            lambda m, t, r: m.damage_costs[t, r]
            == m.resid_damages[t, r] + m.SLR_damages[t, r],
            "damage_costs",
        ),
    )

    return constraints


#################
## Utils
#################


# Damage function


def functional_form(x, m, r, is_slr=False):
    if is_slr:
        form = m.damage_slr_form[r]
        b1, b2, b3 = m.damage_slr_b1[r], m.damage_slr_b2[r], m.damage_slr_b3[r]
        a = m.damage_slr_a[r]
    else:
        form = m.damage_noslr_form[r]
        b1, b2, b3 = m.damage_noslr_b1[r], m.damage_noslr_b2[r], m.damage_noslr_b3[r]
        a = m.damage_noslr_a[r]

    # Linear functional form
    if "Linear" in value(form):
        return a * b1 * x / 100.0

    # Quadratic functional form
    if "Quadratic" in value(form):
        return a * (b1 * x + b2 * x**2) / 100.0

    # Logistic functional form
    if "Logistic" in value(form):
        return a * logistic(x, b1, b2, b3) / 100.0

    raise NotImplementedError


def damage_fct(x, x0, m, r, is_slr):
    damage = functional_form(x, m, r, is_slr)
    if x0 is not None:
        damage -= functional_form(x0, m, r, is_slr)

    return damage


def logistic(x, b1, b2, b3):
    exponent = soft_max(-b3 * x, 10, scale=0.1)  # Avoid exponential overflow
    return b1 / (1 + b2 * exp(exponent)) - b1 / (1 + b2)



"""
Creates a big Plotly figure with each column a region, and these rows:
  Row 1: regional emissions (with baseline) and carbon price
  Row 2: Adapt/damage costs and adaptation levels
  Row 3: GDP (gross/net), consumption, capital stock and utility
  Row 4: Global emissions, temperature and sea level rise
"""

import os
from mimosa.export.utils import Plot


def full_plot(m, filename, include_plotlyjs=True, custom_layout=None):
    if custom_layout is None:
        custom_layout = {}

    plot = Plot(m, horizontal_spacing=0.005, vertical_spacing=0.05)

    plot.add(m.baseline_emissions, row=1, is_fct=True, name="baseline_emissions")
    try:
        plot.add(m.baseline, row=1)
    except AttributeError:
        pass
    plot.add(m.regional_emissions, row=1)
    plot.add(m.carbonprice, row=1, secondary_y=True)
    try:
        plot.add(m.adapt_costs, row=2, stackgroup="costs")
        plot.add(m.resid_damages, row=2, stackgroup="costs")
        plot.add(m.gross_damages, row=2)
    except AttributeError:
        plot.add(m.damage_costs, row=2, stackgroup="costs")
    plot.add(m.rel_abatement_costs, row=2, stackgroup="costs")

    plot.add(m.capital_stock, row=3)
    plot.add(m.GDP_gross, row=3)
    plot.add(m.GDP_net, row=3)
    plot.add(m.consumption, row=3, visible="legendonly")
    plot.add(m.utility, row=3, secondary_y=True, visible="legendonly")
    plot.add(m.abatement_costs, row=3, visible="legendonly")
    plot.add(m.global_emissions, is_regional=False, row=4)
    plot.add(m.temperature, is_regional=False, row=4, secondary_y=True)
    plot.add(m.cumulative_emissions, is_regional=False, row=4, visible="legendonly")

    try:
        plot.add(
            m.global_rel_abatement_costs,
            is_regional=False,
            row=4,
            secondary_y=True,
            visible="legendonly",
        )
    except AttributeError:
        pass

    # Optional values

    try:
        plot.add(
            m.emission_relative_cumulative,
            is_regional=False,
            row=4,
            secondary_y=True,
            visible="legendonly",
        )
        plot.add(
            m.consumption_loss,
            is_regional=False,
            row=4,
            secondary_y=True,
            visible="legendonly",
        )
        plot.add(
            m.smoothed_factor,
            is_regional=False,
            row=4,
            secondary_y=True,
            visible="legendonly",
        )
        plot.add(
            m.netnegative_emissions, is_regional=False, row=4, visible="legendonly"
        )
        plot.add(m.overshoot, is_regional=False, row=4, visible="legendonly")
    except AttributeError:
        pass

    try:
        # RICE
        plot.add(m.adapt_level, row=2, secondary_y=True, visible="legendonly")
        plot.add(m.adapt_FAD, row=2, secondary_y=True, visible="legendonly")
        plot.add(m.adapt_SAD, row=2, secondary_y=True, visible="legendonly")
        plot.add(m.adapt_IAD, row=2, secondary_y=True, visible="legendonly")
        plot.add(m.SLR_damages, row=2, visible="legendonly")
        plot.add(
            m.total_SLR,
            is_regional=False,
            row=4,
            secondary_y=True,
            visible="legendonly",
        )
        plot.add(
            m.SLR, is_regional=False, row=4, secondary_y=True, visible="legendonly"
        )
        plot.add(
            m.CUMGSIC, is_regional=False, row=4, secondary_y=True, visible="legendonly"
        )
        plot.add(
            m.CUMGIS, is_regional=False, row=4, secondary_y=True, visible="legendonly"
        )
        plot.add(
            m.LBD_factor,
            is_regional=False,
            row=4,
            secondary_y=True,
            visible="legendonly",
        )
        plot.add(
            m.test, is_regional=False, row=4, secondary_y=True, visible="legendonly"
        )
    except AttributeError:
        pass

    try:
        # WITCH
        plot.add(m.adapt_costs_reactive, row=2, secondary_y=True, visible="legendonly")
        plot.add(m.adapt_costs_proactive, row=2, secondary_y=True, visible="legendonly")
        plot.add(m.adapt_costs_speccap, row=2, secondary_y=True, visible="legendonly")
        plot.add(m.adapt_Q_cap, row=2, secondary_y=True, visible="legendonly")
        plot.add(m.adapt_K_proactive, row=2, secondary_y=True, visible="legendonly")
    except AttributeError:
        pass

    plot.set_yaxes_titles(
        ["GtCO<sub>2</sub>/yr", "% GDP", "trillion US$2005/yr", "GtCO<sub>2</sub>/yr"]
    )
    plot.fig.update_yaxes(
        title="Carbon price", row=1, col=len(plot.regions), secondary_y=True
    )
    plot.fig.update_yaxes(
        title="Adaptation level", row=2, col=len(plot.regions), secondary_y=True
    )
    plot.fig.update_yaxes(rangemode="tozero", row=2)
    plot.set_layout()
    plot.fig.update_layout(custom_layout)

    outputdir = "output/plots"
    os.makedirs(outputdir + "/", exist_ok=True)
    outputfile = "{}/{}.html".format(outputdir, filename)
    print(f"Plot saved at {outputfile}")
    plot.fig.write_html(outputfile, include_plotlyjs=include_plotlyjs)



"""
Model equations and constraints:
Damage and adaptation costs
"""

from typing import Sequence
from mimosa.common import AbstractModel, Var, GeneralConstraint, RegionalConstraint


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Damage and adaptation costs equations and constraints
    (no-damage specification)

    Necessary variables:
        m.damage_costs (sum of residual damages and adaptation costs, as % of GDP)

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.damage_costs = Var(m.t, m.regions, initialize=0.0)
    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.damage_costs[t, r] == 0.0, "zero_damage_costs"
            )
        ]
    )

    return constraints



"""
Common functions and utilities
"""

import os
import time
import yaml
from mimosa.common import logger


def firstk(dictionary):
    """Returns the first key of a dictionary"""
    keys = list(dictionary.keys())
    return keys[0]


def first(dictionary):
    """Returns the first element (value) of a dictionary"""
    return dictionary[firstk(dictionary)]


def timer(name, log=False):
    """Decorator which times functions

    Arguments:
        name {str} -- Description of the function
    """

    def decorator(fct):
        def wrapper(*args, **kwargs):
            time1 = time.time()
            result = fct(*args, **kwargs)
            time2 = time.time()
            message = "{} took {:.3g} seconds.".format(name, time2 - time1)
            if log:
                logger.info(message)
            print(message)
            return result

        return wrapper

    return decorator


def load_yaml(filename):
    full_filename = os.path.join(
        os.path.dirname(__file__), "../inputdata/config/", filename
    )
    with open(full_filename, "r", encoding="utf8") as configfile:
        output = yaml.safe_load(configfile)
    return output



"""
Model equations and constraints:
Emissions and temperature
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    Param,
    Var,
    GeneralConstraint,
    GlobalConstraint,
    GlobalInitConstraint,
    RegionalConstraint,
    RegionalInitConstraint,
    Constraint,
    value,
    quant,
)


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Emissions and temperature equations and constraints

    Necessary variables:
        m.relative_abatement
        m.cumulative_emissions
        m.T0
        m.temperature

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.regional_emissions = Var(m.t, m.regions, units=quant.unit("emissionsrate_unit"))
    m.baseline = Var(
        m.t,
        m.regions,
        initialize=lambda m, t, r: m.baseline_emissions(m.year(t), r),
        units=quant.unit("emissionsrate_unit"),
    )
    m.baseline_carbon_intensity = Param()

    m.relative_abatement = Var(
        m.t,
        m.regions,
        initialize=0,
        bounds=(0, 2.5),
        units=quant.unit("fraction_of_baseline_emissions"),
    )
    m.regional_emission_reduction = Var(
        m.t, m.regions, units=quant.unit("emissionsrate_unit")
    )
    m.cumulative_emissions = Var(m.t, units=quant.unit("emissions_unit"))
    m.global_emissions = Var(m.t, units=quant.unit("emissionsrate_unit"))

    m.cumulative_emissions_trapz = Param()

    constraints.extend(
        [
            # Baseline emissions based on emissions or carbon intensity
            RegionalConstraint(
                lambda m, t, r: (
                    m.baseline[t, r]
                    == m.carbon_intensity(m.year(t), r) * m.GDP_net[t, r]
                )
                if value(m.baseline_carbon_intensity)
                else (m.baseline[t, r] == m.baseline_emissions(m.year(t), r)),
                name="baseline_emissions",
            ),
            # Regional emissions from baseline and relative abatement
            RegionalConstraint(
                lambda m, t, r: m.regional_emissions[t, r]
                == (1 - m.relative_abatement[t, r])
                * (
                    m.baseline[t, r]
                    if value(m.baseline_carbon_intensity)
                    else m.baseline_emissions(m.year(t), r)
                    # Note: this should simply be m.baseline[t,r], but this is numerically less stable
                    # than m.baseline_emissions(m.year(t), r) whenever baseline intensity
                    # is used instead of baseline emissions. In fact, m.baseline_emissions(m.year(t), r)
                    # is just a fixed number, whereas m.baseline[t,r] is a variable depending on
                    # GDP.
                )
                if t > 0
                else Constraint.Skip,
                "regional_abatement",
            ),
            RegionalInitConstraint(
                lambda m, r: m.regional_emissions[0, r]
                == m.baseline_emissions(m.year(0), r)
            ),
            RegionalConstraint(
                lambda m, t, r: m.regional_emission_reduction[t, r]
                == m.baseline[t, r] - m.regional_emissions[t, r],
                "regional_emission_reduction",
            ),
            # Global emissions (sum from regional emissions)
            GlobalConstraint(
                lambda m, t: m.global_emissions[t]
                == sum(m.regional_emissions[t, r] for r in m.regions)
                if t > 0
                else Constraint.Skip,
                "global_emissions",
            ),
            GlobalInitConstraint(
                lambda m: m.global_emissions[0]
                == sum(m.baseline_emissions(m.year(0), r) for r in m.regions),
                "global_emissions_init",
            ),
            # Cumulative global emissions
            GlobalConstraint(
                lambda m, t: m.cumulative_emissions[t]
                == m.cumulative_emissions[t - 1]
                + (
                    (m.dt * (m.global_emissions[t] + m.global_emissions[t - 1]) / 2)
                    if value(m.cumulative_emissions_trapz)
                    else (m.dt * m.global_emissions[t])
                )
                if t > 0
                else Constraint.Skip,
                "cumulative_emissions",
            ),
            GlobalInitConstraint(lambda m: m.cumulative_emissions[0] == 0),
        ]
    )

    m.T0 = Param(units=quant.unit("degC_above_PI"))
    m.temperature = Var(
        m.t, initialize=lambda m, t: m.T0, units=quant.unit("degC_above_PI")
    )
    m.TCRE = Param()
    m.temperature_target = Param()
    constraints.extend(
        [
            GlobalConstraint(
                lambda m, t: m.temperature[t]
                == m.T0 + m.TCRE * m.cumulative_emissions[t],
                "temperature",
            ),
            GlobalInitConstraint(lambda m: m.temperature[0] == m.T0),
            GlobalConstraint(
                lambda m, t: m.temperature[t] <= m.temperature_target
                if (m.year(t) >= 2100 and value(m.temperature_target) is not False)
                else Constraint.Skip,
                name="temperature_target",
            ),
        ]
    )

    m.perc_reversible_damages = Param()

    # m.overshoot = Var(m.t, initialize=0)
    # m.overshootdot = DerivativeVar(m.overshoot, wrt=m.t)
    # m.netnegative_emissions = Var(m.t)
    # global_constraints.extend(
    #     [
    #         lambda m, t: m.netnegative_emissions[t]
    #         == m.global_emissions[t] * (1 - tanh(m.global_emissions[t] * 10)) / 2
    #         if value(m.perc_reversible_damages) < 1
    #         else Constraint.Skip,
    #         lambda m, t: m.overshootdot[t]
    #         == (m.netnegative_emissions[t] if t <= value(m.year2100) and t > 0 else 0)
    #         if value(m.perc_reversible_damages) < 1
    #         else Constraint.Skip,
    #     ]
    # )

    # global_constraints_init.extend([lambda m: m.overshoot[0] == 0])

    # Emission constraints

    m.budget = Param()
    m.inertia_global = Param()
    m.inertia_regional = Param()
    m.global_min_level = Param()
    m.regional_min_level = Param()
    m.no_pos_emissions_after_budget_year = Param()
    m.non_increasing_emissions_after_2100 = Param()
    constraints.extend(
        [
            # Carbon budget constraints:
            GlobalConstraint(
                lambda m, t: m.cumulative_emissions[t]
                - (
                    m.budget
                    + (
                        m.overshoot[t] * (1 - m.perc_reversible_damages)
                        if value(m.perc_reversible_damages) < 1
                        else 0
                    )
                )
                <= 0
                if (m.year(t) >= 2100 and value(m.budget) is not False)
                else Constraint.Skip,
                name="carbon_budget",
            ),
            GlobalConstraint(
                lambda m, t: m.global_emissions[t] <= 0
                if (
                    m.year(t) >= 2100
                    and value(m.no_pos_emissions_after_budget_year) is True
                    and value(m.budget) is not False
                )
                else Constraint.Skip,
                name="net_zero_after_2100",
            ),
            GlobalConstraint(lambda m, t: m.cumulative_emissions[t] >= 0),
            # Global and regional inertia constraints:
            GlobalConstraint(
                lambda m, t: m.global_emissions[t] - m.global_emissions[t - 1]
                >= m.dt
                * m.inertia_global
                * sum(m.baseline_emissions(m.year(0), r) for r in m.regions)
                if value(m.inertia_global) is not False and t > 0
                else Constraint.Skip,
                name="global_inertia",
            ),
            RegionalConstraint(
                lambda m, t, r: m.regional_emissions[t, r]
                - m.regional_emissions[t - 1, r]
                >= m.dt * m.inertia_regional * m.baseline_emissions(m.year(0), r)
                if value(m.inertia_regional) is not False and t > 0
                else Constraint.Skip,
                name="regional_inertia",
            ),
            RegionalConstraint(
                lambda m, t, r: m.regional_emissions[t, r]
                - m.regional_emissions[t - 1, r]
                <= 0
                if m.year(t - 1) > 2100 and value(m.non_increasing_emissions_after_2100)
                else Constraint.Skip,
                name="non_increasing_emissions_after_2100",
            ),
            GlobalConstraint(
                lambda m, t: m.global_emissions[t] >= m.global_min_level
                if value(m.global_min_level) is not False
                else Constraint.Skip,
                "global_min_level",
            ),
            RegionalConstraint(
                lambda m, t, r: m.regional_emissions[t, r] >= m.regional_min_level
                if value(m.regional_min_level) is not False
                else Constraint.Skip,
                "regional_min_level",
            ),
        ]
    )

    m.emission_relative_cumulative = Var(m.t, initialize=1)
    constraints.extend(
        [
            GlobalConstraint(
                lambda m, t: (
                    m.emission_relative_cumulative[t]
                    == m.cumulative_emissions[t]
                    / m.baseline_cumulative_global(m, m.year(0), m.year(t))
                )
                if t > 0
                else Constraint.Skip,
                name="relative_cumulative_emissions",
            ),
            GlobalInitConstraint(lambda m: m.emission_relative_cumulative[0] == 1),
        ]
    )

    return constraints



"""
Model equations and constraints:
Emission trading module
Type: trading of emissions, and not costs
"""

from typing import Sequence

from mimosa.common import (
    AbstractModel,
    GeneralConstraint,
    GlobalConstraint,
    RegionalConstraint,
    Var,
    NonNegativeReals,
    quant,
    sqrt,
)

from mimosa.components.abatement import AC


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Emission trading equations and constraints
    (trading of emissions specification)

    Necessary variables:
        m.abatement_costs (abatement costs as paid for by this region)

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    # Default options for variables defined below
    opts_emiss = {"units": quant.unit("emissions_unit"), "within": NonNegativeReals}
    opts_costs = {"units": quant.unit("currency_unit"), "within": NonNegativeReals}

    # Emission reduction in region r paid for by region r
    m.own_emission_reduction = Var(m.t, m.regions, **opts_emiss)
    m.own_emission_reduction_costs = Var(m.t, m.regions, **opts_costs)

    # Emission reduction in other regions paid for by region r
    m.imported_emission_reduction = Var(m.t, m.regions, **opts_emiss)
    m.imported_emission_reduction_costs = Var(m.t, m.regions, **opts_costs)

    # Emission reduction in region r paid for by other regions
    m.exported_emission_reduction = Var(m.t, m.regions, initialize=0.01, **opts_emiss)
    m.exported_emission_reduction_costs = Var(m.t, m.regions, **opts_costs)

    # Actual emission reduction (will be equal to own + exported reductions)
    m.actual_emission_reduction = Var(m.t, m.regions, **opts_emiss)

    # Totals of imported and exported emissions
    m.total_exported_reductions = Var(m.t, **opts_emiss)
    m.total_exported_reduction_costs = Var(m.t, **opts_costs)

    constraints.extend(
        [
            GlobalConstraint(
                lambda m, t: m.total_exported_reductions[t]
                == sum(m.exported_emission_reduction[t, r] for r in m.regions),
                "total_exported_emissions",
            ),
            GlobalConstraint(
                lambda m, t: m.total_exported_reduction_costs[t]
                == sum(m.exported_emission_reduction_costs[t, r] for r in m.regions),
                "total_exported_reduction_costs",
            ),
            GlobalConstraint(
                lambda m, t: m.total_exported_reductions[t]
                == sum(m.imported_emission_reduction[t, r] for r in m.regions),
                "total_imported_equals_total_exported_emission_reductions",
            ),
        ]
    )

    constraints.extend(
        [
            # Costs for imported emission reductions: imported reductions have a global price
            RegionalConstraint(
                lambda m, t, r: m.imported_emission_reduction_costs[t, r]
                == m.imported_emission_reduction[t, r]
                * m.total_exported_reduction_costs[t]
                / m.total_exported_reductions[t],
                "costs_for_imported_emission_reductions",
            ),
            # Costs of the own emission reductions: the "cheap" part of the MAC
            RegionalConstraint(
                lambda m, t, r: m.own_emission_reduction_costs[t, r]
                == AC(m.own_emission_reduction[t, r] / m.baseline[t, r], m, t, r)
                * m.baseline[t, r],
                "costs_own_emission_reductions",
            ),
            # Costs of exported emission reductions: calculated as the "expensive" part of the MAC
            # This is the area under the MAC for the total emission reduction (own + exported) minus
            # the costs of the own reductions.
            RegionalConstraint(
                lambda m, t, r: m.actual_emission_reduction[t, r]
                == m.own_emission_reduction[t, r] + m.exported_emission_reduction[t, r],
                "actual_emission_reduction",
            ),
            RegionalConstraint(
                lambda m, t, r: m.exported_emission_reduction_costs[t, r]
                == AC(
                    m.actual_emission_reduction[t, r] / m.baseline[t, r],
                    m,
                    t,
                    r,
                )
                * m.baseline[t, r]
                - m.own_emission_reduction_costs[t, r],
                "costs_exported_emission_reductions",
            ),
            # Actual reductions and actual abatement costs:
            RegionalConstraint(
                lambda m, t, r: m.relative_abatement[t, r]
                == m.actual_emission_reduction[t, r] / m.baseline[t, r],
                "relative_abatement",
            ),
            RegionalConstraint(
                lambda m, t, r: m.abatement_costs[t, r]
                == m.own_emission_reduction_costs[t, r]
                + 1.02 * m.imported_emission_reduction_costs[t, r],
                "abatement_costs",
            ),
        ]
    )

    return constraints



"""
Create a RegionalParamStore object which reads and parses regional
parameter values from various input files

These regional values are:
  [Main]
  - MAC scaling factor
  - Initial capital factor

  [RICE2010]
  - a1, a2, a3
  - g1, g2

  [RICE2012]
  - a1, a2, a3
  - nu1, nu2, nu3,
  - SLRDAM1, SLRDAM2

"""

from .regional_param_container import RegionalParamContainer


###########################
#
# Define the store
#
###########################


class RegionalParamStore:
    """
    Used to read and parse regional parameter values from various input files.

    The data is read from different CSV files: one CSV file per "category". 
    Each category is an object of type `RegionalParameters` and contains
    the data for one or multiple parameters:
        - economics: `init_capital_factor`
        - MAC: `kappa`
        - ADRICE2010: all the regional AD-RICE2010 parameters (`a1`, `a2`, `a3`, `g1`, `g2`)
        - ADRICE2012: all the regional AD-RICE2012 parameters
                        (`a1`, `a2`, `a3`, `nu1`, `nu2`, `nu3`, `slrdam1`, `slrdam2`)
    
    Each of these categories has an associated region type, saying which regions are present.
    Conversion between region types happens within the `RegionalParameters` object.

    To get the parameter values for a parameter name, the function `get` is used (or,
    `getregional` if only a single region is needed)
    """

    def __init__(self, params):
        self.params = params
        self.regions = params["regions"]

        regiontype_output = params["regionstype"]

        self.categories = {
            "economics": RegionalParamContainer(
                "economics.csv", "IMAGE26", regiontype_output
            ),
            "MAC": RegionalParamContainer("mac.csv", "IMAGE26", regiontype_output),
            "ADRICE2010": RegionalParamContainer(
                "ADRICE2010.csv", "ADRICE2010", regiontype_output
            ),
            "ADRICE2012": RegionalParamContainer(
                "ADRICE2012.csv", "ADRICE2012", regiontype_output
            ),
            "COACCH": RegionalParamContainer("COACCH.csv", "COACCH", regiontype_output),
        }

    def get(self, category, paramname):
        return {
            region: self.getregional(category, paramname, region)
            for region in self.regions
        }

    def getregional(self, category, paramname, region):
        # First check if parameter is set manually in config file
        try:
            value = self.params["regions"][region][category][paramname]
        except (KeyError, TypeError):
            value = self.categories[category].get(paramname, region)

        return value



"""
The simulation mode imposes exogenously an emission path, carbon price or
other regional or global variable.

The extra constraints are defined in the params dict, under:
    params["simulation"]["constraint_variables"] = {
        'variablename1': 'path_of_outputfile_to_be_used.csv',
        ...
    }

At the same time, certain constraints can be deactivated. These are defined in:
    params["simulation"]["disabled_constraints"] = ['carbon_budget', ...]

"""

from mimosa.common import ConcreteModel
from .deactivate_constraints import deactivate_constraints
from .set_constraints import set_constraints_fixed_variables


def set_simulation_mode(m: ConcreteModel, params: dict) -> None:
    # Add constraints for each fixed variable
    set_constraints_fixed_variables(m, params)

    # Disable constraints to allow feasible solution
    deactivate_constraints(m, params)



"""
Model equations and constraints:
Damage and adaptation costs, WITCH specification
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    Param,
    Var,
    GeneralConstraint,
    RegionalConstraint,
    RegionalInitConstraint,
)


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Damage and adaptation costs equations and constraints
    (WITCH specification)

    Necessary variables:
        m.damage_costs (sum of residual damages and adaptation costs, % of GDP)

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.damage_costs = Var(m.t, m.regions)

    # Gross and residual damages
    m.gross_damages = Var(m.t, m.regions)
    m.resid_damages = Var(m.t, m.regions)
    m.adapt_Q_ada = Var(m.t, m.regions, initialize=0.1)

    m.damage_omega1_pos = Param(m.regions)
    m.damage_omega1_neg = Param(m.regions)
    m.damage_omega2_pos = Param(m.regions)
    m.damage_omega2_neg = Param(m.regions)
    m.damage_omega3_pos = Param(m.regions)
    m.damage_omega3_neg = Param(m.regions)
    m.damage_omega4_pos = Param(m.regions)
    m.damage_omega4_neg = Param(m.regions)

    m.adapt_eps = Param(m.regions)

    m.damage_scale_factor = Param()

    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.gross_damages[t, r]
                == m.damage_scale_factor * damage_fct(m.temperature[t], 0, m.T0, m, r),
                "gross_damages",
            ),
            RegionalConstraint(
                lambda m, t, r: m.resid_damages[t, r]
                == m.damage_scale_factor
                * damage_fct(
                    m.temperature[t],
                    pow(m.adapt_Q_ada[t, r], m.adapt_eps[r], True),
                    m.T0,
                    m,
                    r,
                ),
                "resid_damages",
            ),
        ]
    )

    # Adaptation costs (protection costs, PC)
    m.adapt_costs = Var(m.t, m.regions)
    m.adapt_costs_reactive = Var(m.t, m.regions, bounds=(0, 1))  # I_RADA
    # m.adapt_costs_proactive   = Var(m.t, m.regions, bounds=(0,1)) # I_PRADA
    # m.adapt_costs_speccap     = Var(m.t, m.regions) # I_SCAP, specific adaptive capacity
    m.fixed_adaptation = Param()

    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.adapt_costs[t, r]
                == (
                    m.adapt_costs_reactive[t, r]  # +
                    # m.adapt_costs_proactive[t,r] +
                    # m.adapt_costs_speccap[t,r]
                ),
                name="adapt_costs",
            ),
            RegionalInitConstraint(lambda m, r: m.adapt_costs_reactive[0, r] == 0),
        ]
    )

    ## Nested Constant Elasticity of Substitution (CES)

    m.adapt_omega_eff_ada = Param(m.regions)
    m.adapt_omega_act = Param(m.regions)
    m.adapt_omega_eff_act = Param(m.regions)
    m.adapt_omega_rada = Param(m.regions)
    m.adapt_rho_ada = Param(m.regions)
    m.adapt_rho_act = Param(m.regions)

    # Nest 1: adaptation activities vs adaptive capacity
    # m.adapt_Q_act = Var(m.t, m.regions, initialize=0.1)
    m.adapt_Q_cap = Var(m.t, m.regions, initialize=0.1)

    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.adapt_Q_ada[t, r]
                == m.adapt_omega_eff_ada[r]
                * m.adapt_omega_eff_act[r]
                * m.adapt_costs_reactive[t, r],
                "adapt_Q_ada",
            ),
            # Extra constraint necessary to avoid negative GDP
            RegionalConstraint(lambda m, t, r: m.GDP_net[t, r] >= 0),
        ]
    )

    # regional_constraints.append(lambda m,t,r: m.adapt_Q_ada[t,r] == CES(
    #     m.adapt_costs_reactive[t,r], m.adapt_Q_cap[t,r],
    #     # m.adapt_Q_act[t,r], m.adapt_Q_cap[t,r],
    #     m.adapt_omega_eff_ada[r], m.adapt_omega_act[r], m.adapt_rho_ada[r]
    # ))

    # Nest 2: reactive adaptation vs proactive adaptation
    # m.adapt_K_proactive = Var(m.t, m.regions, initialize=0.1)
    # regional_constraints.append(lambda m,t,r: m.adapt_Q_act[t,r] == CES(
    #     m.adapt_costs_reactive[t,r], m.adapt_K_proactive[t,r],
    #     m.adapt_omega_eff_act[r], m.adapt_omega_rada[r], m.adapt_rho_act[r]
    # ))

    # Nest 3: we don't model R&D and human capital, so we use the description
    # of Bosello et al (2010) [original AD-WITCH] that Q_cap is simple stock variable

    # Stocks of capital
    # m.adapt_Q_capdot        = DerivativeVar(m.adapt_Q_cap, wrt=m.t)
    # m.adapt_K_proactivedot  = DerivativeVar(m.adapt_K_proactive, wrt=m.t)

    # regional_constraints.extend([
    #     lambda m,t,r: m.adapt_Q_capdot[t,r]         == np.log(1-m.dk) * m.adapt_Q_cap[t,r] + m.adapt_costs_speccap[t,r],
    #     # lambda m,t,r: m.adapt_K_proactivedot[t,r]   == np.log(1-m.dk) * m.adapt_K_proactive[t,r] + m.adapt_costs_proactive[t,r],
    # ])

    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.damage_costs[t, r]
                == m.resid_damages[t, r] + m.adapt_costs[t, r],
                "damage_costs",
            )
        ]
    )

    return constraints


#################
## Utils
#################


# Damage function


def damage_fct(T, Q_adapt, T0, m, r):
    return _damage_fct(
        T,
        Q_adapt,
        T0,
        m.damage_omega1_pos[r],
        m.damage_omega1_neg[r],
        m.damage_omega2_pos[r],
        m.damage_omega2_neg[r],
        m.damage_omega3_pos[r],
        m.damage_omega3_neg[r],
        m.damage_omega4_pos[r],
        m.damage_omega4_neg[r],
    )


def damage_fct_dot(T, Q_adapt, m, r):
    raise NotImplementedError


def _damage_fct(
    T, Q_adapt, T0, w1_pos, w1_neg, w2_pos, w2_neg, w3_pos, w3_neg, w4_pos, w4_neg
):
    """Damage function

    T: temperature
    T0 [None]: if specified, substracts damage at T0
    """
    fct = lambda temp, adapt: (
        (w1_neg * temp + w2_neg * temp**w3_neg + w4_neg) / (1 + adapt)
        + (w1_pos * temp + w2_pos * temp**w3_pos + w4_pos)
    )
    dmg = fct(T, Q_adapt)
    if T0 is not None:
        dmg -= fct(T0, 0)
    return dmg


# Constant Elasticity of Substitution
def CES(Q1, Q2, omega_eff, omega, rho):
    return omega_eff * (omega * Q1**rho + (1 - omega) * Q2**rho) ** (1 / rho)



def get_nested(dictionary, keys, create=False):
    for key in keys:
        if key not in dictionary and create:
            dictionary[key] = {}
        dictionary = dictionary[key]
    return dictionary


def set_nested(dictionary, keys, value):
    dictionary = get_nested(dictionary, keys[:-1], True)
    dictionary[keys[-1]] = value


def flatten(joined_dict, leaf_criterium=None):

    flattened_dict = {}

    leaf_criterium_default = lambda keys, node: not isinstance(node, dict)

    if leaf_criterium is None:
        leaf_criterium = leaf_criterium_default

    def _recursive_traverse(curr_keys, subset):
        for key, node in subset.items():
            keys = list(curr_keys) + [key]
            if leaf_criterium(keys, node):
                flattened_dict[" - ".join(keys)] = node
            else:
                _recursive_traverse(keys, node)

    _recursive_traverse([], joined_dict)

    return flattened_dict



from mimosa import MIMOSA, load_params

params = load_params()

# Make changes to the params if needed
params["emissions"]["carbonbudget"] = False

model1 = MIMOSA(params)
model1.solve()
model1.save("run1")



"""
Model equations and constraints:
Sea level rise (height of sea level rise, not SLR damages)
From RICE 2010
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    Param,
    Var,
    GeneralConstraint,
    GlobalConstraint,
    GlobalInitConstraint,
    Constraint,
    NonNegativeReals,
    quant,
)


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Emissions and temperature equations and constraints

    Necessary variables:
        m.total_SLR

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """

    # Parameters and variables necessary for sea level rise
    m.slr_thermal = Var(m.t, within=NonNegativeReals, units=quant.unit("m"))
    m.slr_thermal_equil = Param()
    m.slr_thermal_init = Param()
    m.slr_thermal_adjust_rate = Param()

    m.slr_cumgsic = Var(m.t, within=NonNegativeReals, units=quant.unit("m"))
    m.slr_gsic_melt_rate = Param()
    m.slr_gsic_total_ice = Param()
    m.slr_gsic_equil_temp = Param()

    m.slr_cumgis = Var(m.t, within=NonNegativeReals, units=quant.unit("m"))
    m.slr_gis_melt_rate_above_thresh = Param()
    m.slr_gis_init_melt_rate = Param()
    m.slr_gis_init_ice_vol = Param()

    m.total_SLR = Var(m.t, within=NonNegativeReals, units=quant.unit("m"))

    # Constraints relating to SLR
    constraints = [
        # Thermal expansion
        GlobalConstraint(
            lambda m, t: m.slr_thermal[t]
            == slr_thermal_expansion(m.slr_thermal[t - 1], m.temperature[t - 1], m)
            if t > 0
            else Constraint.Skip,
            name="SLR_thermal",
        ),
        GlobalInitConstraint(
            lambda m: m.slr_thermal[0] == slr_thermal_expansion_init(m)
        ),
        # GSIC
        GlobalConstraint(
            lambda m, t: m.slr_cumgsic[t]
            == slr_gsic(m.slr_cumgsic[t - 1], m.temperature[t - 1], m)
            if t > 0
            else Constraint.Skip,
            name="SLR_GSIC",
        ),
        GlobalInitConstraint(lambda m: m.slr_cumgsic[0] == 0.015),
        # GIS
        GlobalConstraint(
            lambda m, t: m.slr_cumgis[t]
            == slr_gis(m.slr_cumgis[t - 1], m.temperature[t - 1], m)
            if t > 0
            else Constraint.Skip,
            name="SLR_GIS",
        ),
        GlobalInitConstraint(lambda m: m.slr_cumgis[0] == 0.006),
        # Total SLR is sum of each contributing factors
        GlobalConstraint(
            lambda m, t: m.total_SLR[t]
            == m.slr_thermal[t] + m.slr_cumgsic[t] + m.slr_cumgis[t],
            name="total_SLR",
        ),
    ]

    return constraints


def slr_thermal_expansion_init(m):
    """Calculates initial SLR thermal expansion"""
    return m.slr_thermal_init + m.slr_thermal_adjust_rate * (
        m.T0 * m.slr_thermal_equil - m.slr_thermal_init
    )


def slr_thermal_expansion(slr_thermal, temperature, m: AbstractModel):
    """Calculates next-step SLR due to thermal expansian"""

    equilib = m.slr_thermal_equil
    adjust_rate = m.slr_thermal_adjust_rate

    return (1 - adjust_rate) ** (m.dt / 10) * slr_thermal + adjust_rate * (
        m.dt / 10
    ) * (temperature * equilib)


def slr_gsic(cumgsic, temperature, m: AbstractModel):
    """Calculates next-step SLR due to glaciers and small ice caps"""

    melt_rate = m.slr_gsic_melt_rate
    total_ice = m.slr_gsic_total_ice
    equil_temp = m.slr_gsic_equil_temp

    return cumgsic + melt_rate / total_ice * m.dt * (total_ice - cumgsic) * (
        temperature - equil_temp
    )


def slr_gis(cumgis, temperature, m: AbstractModel):
    """Calculates next-step SLR due to the Greenland ice sheet"""

    melt_rate_above_thresh = m.slr_gis_melt_rate_above_thresh
    init_melt_rate = m.slr_gis_init_melt_rate
    init_ice_vol = m.slr_gis_init_ice_vol

    return cumgis + (m.dt / 10) * (1 / 100) * (
        melt_rate_above_thresh * temperature + init_melt_rate
    ) * (1 - cumgis / init_ice_vol)



"""
Model equations and constraints:
Objective function
"""

from typing import Sequence, Tuple

from mimosa.common import (
    AbstractModel,
    Constraint,
    GeneralConstraint,
    GlobalConstraint,
    GlobalInitConstraint,
    Objective,
    Param,
    Var,
    exp,
    minimize,
)


def get_constraints(m: AbstractModel) -> Tuple[Objective, Sequence[GeneralConstraint]]:
    """Equations and constraints for the objective of the optimisation
    (global costs specification)

    Necessary variables:


    Returns:
        - Objective
        - list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.NPV = Var(m.t)
    m.PRTP = Param()
    constraints.extend(
        [
            GlobalConstraint(
                lambda m, t: m.NPV[t]
                == m.NPV[t - 1]
                + m.dt
                * exp(-m.PRTP * (m.year(t) - m.beginyear))
                * (
                    sum(m.abatement_costs[t, r] for r in m.regions)
                    + sum(m.damage_costs[t, r] * m.GDP_gross[t, r] for r in m.regions)
                )
                if t > 0
                else Constraint.Skip,
                name="NPV",
            ),
            GlobalInitConstraint(lambda m: m.NPV[0] == 0),
        ]
    )

    objective = Objective(rule=lambda m: m.NPV[m.tf], sense=minimize)

    return objective, constraints



"""
Add constraints for each fixed variable
"""

import pandas as pd
from mimosa.common import (
    RegionalConstraint,
    GlobalConstraint,
    is_regional,
    add_constraint,
)
from .utils import InterpolatingData, read_csv


def set_constraints_fixed_variables(m, params):
    data_cache = {}
    extra_constraints = []

    for variable_name, filepath_or_data in params["simulation"][
        "constraint_variables"
    ].items():
        interp_data = _get_interp_data(filepath_or_data, data_cache, variable_name)
        extra_constraints.extend(_fixed_data_constraint(m, variable_name, interp_data))

    # Add constraints to concrete model
    for constraint in extra_constraints:
        add_constraint(m, constraint.to_pyomo_constraint(m), constraint.name)


def _get_interp_data(filepath_or_data, data_cache, variable_name):
    if isinstance(filepath_or_data, dict):
        data = pd.DataFrame(filepath_or_data)
        # Important: data should be a dataframe with years as columns and regions as row indices
        return InterpolatingData(data)

    # If it is a string, read the data from the CSV file
    filepath = filepath_or_data
    # 1. Read datafile
    if filepath not in data_cache:
        data_cache[filepath] = read_csv(filepath)
    imported_data = data_cache[filepath]

    # 2. Select data
    fixed_data = imported_data[imported_data["Variable"] == variable_name]
    fixed_data = fixed_data.drop(
        columns=["Variable", "Unit"], errors="ignore"
    ).set_index("Region")
    interp_data = InterpolatingData(fixed_data)
    return interp_data


def _fixed_data_constraint(m, variable_name, interp_data):
    # 3. Create constraints out of this
    eps = 1e-3  # TODO make eps a variable

    if is_regional(getattr(m, variable_name)):
        extra_constraints = _extra_regional_constraint(variable_name, interp_data, eps)
    else:
        extra_constraints = _extra_global_constraint(variable_name, interp_data, eps)

    return extra_constraints


###########
#
# Make regional or global constraints
#
############


def _extra_regional_constraint(variable_name, interp_data, eps):
    return [
        RegionalConstraint(
            lambda m, t, r: getattr(m, variable_name)[t, r]
            - interp_data.get(r, m.year(t))
            <= eps
            # if m.year(t) <= interp_data.maxyear
            # else Constraint.Skip
        ),
        RegionalConstraint(
            lambda m, t, r: getattr(m, variable_name)[t, r]
            - interp_data.get(r, m.year(t))
            >= -eps
            # if m.year(t) <= interp_data.maxyear
            # else Constraint.Skip
        ),
    ]


def _extra_global_constraint(variable_name, interp_data: InterpolatingData, eps):
    return [
        GlobalConstraint(
            lambda m, t: getattr(m, variable_name)[t]
            - interp_data.get("Global", m.year(t))
            <= eps
            # if t > 0  # and m.year(t) <= interp_data.maxyear
            # else Constraint.Skip
        ),
        GlobalConstraint(
            lambda m, t: getattr(m, variable_name)[t]
            - interp_data.get("Global", m.year(t))
            >= -eps
            # if t > 0  # and m.year(t) <= interp_data.maxyear
            # else Constraint.Skip
        ),
    ]



"""
Model equations and constraints:
Abatement costs
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    Param,
    Var,
    GeneralConstraint,
    GlobalConstraint,
    RegionalConstraint,
    RegionalInitConstraint,
    Constraint,
    value,
    log,
    soft_min,
    NonNegativeReals,
    quant,
)


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Abatement cost equations and constraints

    Necessary variables:
        m.abatement_costs

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    ### Technological learning

    # Learning by doing
    m.LBD_rate = Param()
    m.log_LBD_rate = Param(initialize=log(m.LBD_rate) / log(2))
    m.LBD_scaling = Param()
    m.LBD_factor = Var(m.t)  # , bounds=(0,1), initialize=1)
    constraints.append(
        GlobalConstraint(
            lambda m, t: m.LBD_factor[t]
            == soft_min(
                (
                    m.baseline_cumulative_global(m, m.year(0), m.year(t))
                    - m.cumulative_emissions[t]
                )
                / m.LBD_scaling
                + 1.0
            )
            ** m.log_LBD_rate,
            name="LBD",
        )
    )

    # Learning over time and total learning factor
    m.LOT_rate = Param()
    m.LOT_factor = Var(m.t)
    m.learning_factor = Var(m.t)
    constraints.extend(
        [
            GlobalConstraint(
                lambda m, t: m.LOT_factor[t] == 1 / (1 + m.LOT_rate) ** t, "LOT"
            ),
            GlobalConstraint(
                lambda m, t: m.learning_factor[t]
                == (m.LBD_factor[t] * m.LOT_factor[t]),
                "learning",
            ),
        ]
    )

    # Abatement costs and MAC
    m.abatement_costs = Var(
        m.t,
        m.regions,
        # within=NonNegativeReals,
        initialize=0,
        units=quant.unit("currency_unit"),
    )
    m.rel_abatement_costs = Var(m.t, m.regions, units=quant.unit("fraction_of_GDP"))
    m.MAC_gamma = Param()
    m.MAC_beta = Param()
    m.MAC_scaling_factor = Param(m.regions)  # Regional scaling of the MAC
    m.carbonprice = Var(
        m.t,
        m.regions,
        bounds=lambda m: (0, 2 * m.MAC_gamma),
        units=quant.unit("currency_unit/emissions_unit"),
    )
    m.rel_abatement_costs_min_level = Param()
    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.rel_abatement_costs[t, r]
                == m.abatement_costs[t, r] / m.GDP_gross[t, r],
                "rel_abatement_costs",
            ),
            RegionalConstraint(
                lambda m, t, r: m.rel_abatement_costs[t, r]
                >= (m.rel_abatement_costs_min_level if t > 0 else 0.0),
                "rel_abatement_costs_non_negative",
            ),
            RegionalConstraint(
                lambda m, t, r: m.carbonprice[t, r]
                == MAC(m.relative_abatement[t, r], m, t, r),
                "carbonprice",
            ),
            RegionalInitConstraint(
                lambda m, r: m.carbonprice[0, r] == 0, "init_carbon_price"
            ),
        ]
    )

    # Keep track of relative global costs
    m.global_rel_abatement_costs = Var(m.t)
    constraints.extend(
        [
            GlobalConstraint(
                lambda m, t: m.global_rel_abatement_costs[t]
                == sum(m.abatement_costs[t, r] for r in m.regions)
                / sum(m.GDP_gross[t, r] for r in m.regions),
                "global_rel_abatement_costs",
            )
        ]
    )

    # Calculate average global emission reduction per cost unit
    # and average cost per unit emission reduction

    m.global_emission_reduction_per_cost_unit = Var(
        m.t, units=quant.unit("emissionsrate_unit / currency_unit")
    )
    m.global_cost_per_emission_reduction_unit = Var(
        m.t, units=quant.unit("currency_unit / emissionsrate_unit")
    )
    constraints.extend(
        [
            GlobalConstraint(
                lambda m, t: m.global_emission_reduction_per_cost_unit[t]
                == sum(m.regional_emission_reduction[t, r] for r in m.regions)
                / soft_min(sum(m.abatement_costs[t, r] for r in m.regions))
                if t > 0
                else Constraint.Skip,
                "global_emission_reduction_per_cost_unit",
            ),
            GlobalConstraint(
                lambda m, t: m.global_cost_per_emission_reduction_unit[t]
                == sum(m.abatement_costs[t, r] for r in m.regions)
                / soft_min(sum(m.regional_emission_reduction[t, r] for r in m.regions))
                if t > 0
                else Constraint.Skip,
                "global_cost_per_emission_reduction_unit",
            ),
        ]
    )

    return constraints


#################
## Utils
#################


def MAC(a, m, t, r):
    factor = m.learning_factor[t] * m.MAC_scaling_factor[r]
    return factor * m.MAC_gamma * a**m.MAC_beta


def AC(a, m, t, r):
    factor = m.learning_factor[t] * m.MAC_scaling_factor[r]
    return factor * m.MAC_gamma * a ** (m.MAC_beta + 1) / (m.MAC_beta + 1)



"""
Model equations and constraints:
Damage and adaptation costs, RICE specification
"""

from typing import Sequence
from mimosa.common import (
    AbstractModel,
    Param,
    Var,
    GeneralConstraint,
    RegionalConstraint,
    RegionalInitConstraint,
    Constraint,
    soft_min,
    NonNegativeReals,
)


def get_constraints(m: AbstractModel) -> Sequence[GeneralConstraint]:
    """Damage and adaptation costs equations and constraints
    (RICE specification)

    Necessary variables:
        m.damage_costs (sum of residual damages and adaptation costs multiplied by gross GDP)

    Returns:
        list of constraints (any of:
           - GlobalConstraint
           - GlobalInitConstraint
           - RegionalConstraint
           - RegionalInitConstraint
        )
    """
    constraints = []

    m.damage_costs = Var(m.t, m.regions)
    m.gross_damages = Var(m.t, m.regions)
    m.resid_damages = Var(m.t, m.regions)

    m.damage_a1 = Param(m.regions)
    m.damage_a2 = Param(m.regions)
    m.damage_a3 = Param(m.regions)
    m.damage_scale_factor = Param()

    m.adapt_level = Var(m.t, m.regions, within=NonNegativeReals)
    m.adapt_costs = Var(m.t, m.regions)
    m.adapt_FAD = Var(m.t, m.regions, bounds=(0, 0.15), initialize=0)
    m.adapt_IAD = Var(m.t, m.regions, bounds=(0, 0.15), initialize=0)
    m.adap1 = Param(m.regions)
    m.adap2 = Param(m.regions)
    m.adap3 = Param(m.regions)
    m.adapt_rho = Param()
    m.fixed_adaptation = Param()

    m.adapt_SAD = Var(m.t, m.regions, initialize=0.01, within=NonNegativeReals)

    # SLR damages
    m.SLRdam1 = Param(m.regions)
    m.SLRdam2 = Param(m.regions)
    m.SLR_damages = Var(m.t, m.regions)

    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.SLR_damages[t, r]
                == slr_damages(
                    m.total_SLR[t],
                    m.GDP_gross[t, r],
                    m.GDP_gross[0, r],
                    m.SLRdam1[r],
                    m.SLRdam2[r],
                ),
                "SLR_damages",
            ),
        ]
    )

    # Gross damages and adaptation levels
    constraints.extend(
        [
            RegionalConstraint(
                lambda m, t, r: m.gross_damages[t, r]
                == m.damage_scale_factor * damage_fct(m.temperature[t], m.T0, m, r),
                "gross_damages",
            ),
            RegionalConstraint(
                lambda m, t, r: m.resid_damages[t, r]
                == m.gross_damages[t, r] / (1 + m.adapt_level[t, r])
                + m.SLR_damages[t, r],
                "resid_damages",
            ),
            RegionalConstraint(
                lambda m, t, r: m.adapt_SAD[t, r]
                == (1 - m.dk) ** m.dt * m.adapt_SAD[t - 1, r] + m.adapt_IAD[t, r]
                if t > 0
                else Constraint.Skip,
                "adapt_SAD",
            ),
            RegionalInitConstraint(lambda m, r: m.adapt_SAD[0, r] == 0),
            RegionalInitConstraint(lambda m, r: m.adapt_IAD[0, r] == 0),
            RegionalInitConstraint(lambda m, r: m.adapt_FAD[0, r] == 0),
            RegionalConstraint(
                lambda m, t, r: (
                    m.adapt_level[t, r]
                    == m.adap1[r]
                    * (
                        m.adap2[r]
                        * soft_min(m.adapt_FAD[t, r], scale=0.005) ** m.adapt_rho
                        + (1 - m.adap2[r])
                        * soft_min(m.adapt_SAD[t, r], scale=0.005) ** m.adapt_rho
                    )
                    ** (m.adap3[r] / m.adapt_rho)
                )
                if t > 0
                else (m.adapt_level[t, r] == 0),
                name="adapt_level",
            ),
            RegionalConstraint(
                lambda m, t, r: m.adapt_costs[t, r]
                == m.adapt_FAD[t, r] + m.adapt_IAD[t, r],
                "adapt_costs",
            ),
            RegionalConstraint(
                lambda m, t, r: m.damage_costs[t, r]
                == m.resid_damages[t, r] + m.adapt_costs[t, r],
                "damage_costs",
            ),
        ]
    )

    return constraints


#################
## Utils
#################


# Damage function


def damage_fct(temperature, init_temp, m, r):
    power_fct = (
        lambda temp: m.damage_a1[r] * temp + m.damage_a2[r] * temp ** m.damage_a3[r]
    )

    damage = power_fct(soft_min(temperature))
    if init_temp is not None:
        damage -= power_fct(init_temp)

    return damage


# Sea level rise


def slr_damages(total_slr, gdp_t, gdp_0, dam1, dam2):
    return (
        4 * (dam1 * total_slr + dam2 * total_slr**2) * soft_min(gdp_t / gdp_0) ** 0.25
    )



import typing
import re
from abc import ABC, abstractmethod
import numpy as np
from pyomo.core.base.units_container import PintUnitExtractionVisitor

from pyomo.environ import Var, Constraint
import pyomo.environ

# Monkey patch to make sure that the unit stays the same after calling a arctan-function.
# By default, the unit becomes "radian", causing trouble in subsequent unit comparisons.
# Since we use arctan a lot as a "soft-min/max" function, the unit should stay the same as
# the input unit
PintUnitExtractionVisitor.unary_function_method_map[
    "atan"
] = PintUnitExtractionVisitor._get_unit_for_single_child


from pyomo.environ import units


def atan(x):
    name = type(x).__name__
    if name in ["Series", "DataFrame", "float", "int", "ndarray"]:
        return np.arctan(x)

    return pyomo.environ.atan(x)


def exp(x):
    name = type(x).__name__
    if name in ["Series", "DataFrame", "float", "int", "ndarray"]:
        return np.exp(x)

    return pyomo.environ.exp(x)


####### Extra functions


def scale_to_a(scale):
    return 25.0 / scale


def soft_switch(x, scale=1.0):
    """Approximates 0 for x <= 0 and 1 for x > 0

    Args:
        x
        scale (float, optional): order of magnitude of expected values. Defaults to 1.0.
    """
    a = scale_to_a(scale)
    return atan(a * x) / np.pi + 0.5


def soft_min(x, scale=1.0):
    """Soft minimum: approximates the function f(x)=x for x > 0 and f(x)=0 for x <= 0

    Args:
        x
        scale (float, optional): order of magnitude of expected values. Defaults to 1.0.

    Returns:
        approximately x if x > 0 and 0 if x <= 0
    """
    a = scale_to_a(scale)
    soft_min_value = soft_switch(x, scale) * x + 1 / (a * np.pi)
    return soft_min_value

    # Make sure the final answer is positive. Increases computing time, but reduces errors.
    # return sqrt(
    #     soft_min_value ** 2
    # )


def soft_max(x, maxval, scale=1.0):
    return -soft_min(maxval - x, scale) + maxval


####### Constraints


class GeneralConstraint(ABC):
    def __init__(self, rule: typing.Callable, name: str = None):
        """Adds a constraint to the Pyomo mimosa.

        Args:
            rule (typing.Callable): function with parameters m, [t], [r]
            name (str, optional): name of the constraint, useful for debugging. Defaults to None.
        """

        self.name = name
        self.rule = rule

    @abstractmethod
    def to_pyomo_constraint(self, m):
        pass


class GlobalConstraint(GeneralConstraint):
    def to_pyomo_constraint(self, m):
        return Constraint(m.t, rule=self.rule)


class GlobalInitConstraint(GeneralConstraint):
    def to_pyomo_constraint(self, m):
        return Constraint(rule=self.rule)


class RegionalConstraint(GeneralConstraint):
    def to_pyomo_constraint(self, m):
        return Constraint(m.t, m.regions, rule=self.rule)


class RegionalInitConstraint(GeneralConstraint):
    def to_pyomo_constraint(self, m):
        return Constraint(m.regions, rule=self.rule)


class GeneralSoftEqualityConstraint(GeneralConstraint):
    def __init__(
        self,
        rule_lhs: typing.Callable,
        rule_rhs: typing.Callable,
        name: str = None,
        epsilon: float = 0.005,
        absolute_epsilon: float = None,
        ignore_if: typing.Callable = None,
    ):
        """Creates a constraint of the type:
            rule_lhs(x) <= (1 + eps) * rule_rhs(x) && rule_lhs(x) >= (1 - eps) * rule_rhs(x)
        Useful for creating "soft-equality" constraints

        Args:
            rule_lhs (typing.Callable): function (left-hand side) with parameters m, [t], [r]
            rule_rhs (typing.Callable): function (right-hand side) with parameters m, [t], [r]
            name (str, optional): name of the constraint, useful for debugging. Defaults to None.
            epsilon (float, optional): tolerance for upper/lower bounds. Defaults to 0.005.
            ignore_if (typing.Callable): function with parameters m that returns True when constraint should be ignored
        """
        super().__init__(rule_lhs, [f"{name}_upperbound", f"{name}_lowerbound"])

        self.rule_rhs = rule_rhs
        self.epsilon = epsilon
        self.absolute_epsilon = absolute_epsilon

        if ignore_if is None:
            ignore_if = lambda m: False  # Never ignore when ignore_if is None
        self.ignore_if = ignore_if

    def rhs_eps(self, rule_rhs, is_upper: bool, *args):
        factor = 1 if is_upper else -1
        if self.absolute_epsilon is not None:
            return rule_rhs(*args) + factor * self.absolute_epsilon
        return (1 + factor * self.epsilon) * rule_rhs(*args)


class GlobalSoftEqualityConstraint(GeneralSoftEqualityConstraint):
    def to_pyomo_constraint(self, m):
        eps = self.epsilon
        rule_lhs = self.rule
        rule_rhs = self.rule_rhs

        upperbound = (
            lambda m, t: rule_lhs(m, t) <= self.rhs_eps(rule_rhs, True, m, t)
            if not self.ignore_if(m, t)
            else Constraint.Skip
        )
        lowerbound = (
            lambda m, t: rule_lhs(m, t) >= self.rhs_eps(rule_rhs, False, m, t)
            if not self.ignore_if(m, t)
            else Constraint.Skip
        )
        return [
            Constraint(m.t, rule=upperbound),
            Constraint(m.t, rule=lowerbound),
        ]


class RegionalSoftEqualityConstraint(GeneralSoftEqualityConstraint):
    def to_pyomo_constraint(self, m):
        eps = self.epsilon
        rule_lhs = self.rule
        rule_rhs = self.rule_rhs

        upperbound = (
            lambda m, t, r: rule_lhs(m, t, r) <= self.rhs_eps(rule_rhs, True, m, t, r)
            if not self.ignore_if(m, t, r)
            else Constraint.Skip
        )
        lowerbound = (
            lambda m, t, r: rule_lhs(m, t, r) >= self.rhs_eps(rule_rhs, False, m, t, r)
            if not self.ignore_if(m, t, r)
            else Constraint.Skip
        )
        return [
            Constraint(m.t, m.regions, rule=upperbound),
            Constraint(m.t, m.regions, rule=lowerbound),
        ]


def add_constraint(m, constraints, names=None):
    """Adds a constraint to the model

    It first generates a unique name, then adds
    the constraint using this new name
    """
    if not isinstance(constraints, list):
        constraints = [constraints]
        names = [names]

    for constraint, name in zip(constraints, names):
        n = len(list(m.component_objects()))
        name = f"constraint_{n}" if name is None else f"constraint_{name}"
        m.add_component(name, constraint)
    return name


####### Get all variables of a model


class UsefulVar:
    def __init__(self, m, name: str):
        self.m = m
        self.var = getattr(m, name)

        self.name = name
        self.is_regional = is_regional(self.var)
        self.unit = get_unit(self.var)
        self.indices = get_indices(self.var)

        self.index_values = {index: list(getattr(m, index)) for index in self.indices}


def get_all_variables(m):
    return [
        UsefulVar(m, var.name)
        for var in m.component_objects(Var)
        if not var.name.startswith("_")
    ]


def is_regional(var):
    """Returns true if the Pyomo variable `var` is regional, false if it is global"""
    # While there is no explicit Pyomo way to obtain the indices, we can use
    # this private property to check if variable has multiple indices
    if var._implicit_subsets is None:
        return False
    return True


def get_indices(var):
    if is_regional(var):
        return [index.name for index in var._implicit_subsets]
    return [var.index_set().name]


def get_unit(var):
    pyomo_unit = units.get_units(var)
    pyomo_unit_str = str(pyomo_unit) if pyomo_unit is not None else ""

    # Bugfix replace "/a" to "/yr" ("annum" is less clear than year)
    # Note that we should not replace e.g. "/atm", hence the negative lookahead
    return re.sub("/a(?![a-zA-Z])", "/yr", pyomo_unit_str)



import numpy as np
from mimosa.common import AbstractModel, quant
from mimosa.common.data import DataStore
from mimosa.common.regional_params import RegionalParamStore

# Util to create a None-indexed dictionary for scalar components
# (see https://pyomo.readthedocs.io/en/stable/working_abstractmodels/data/raw_dicts.html)
V = lambda val: {None: val}


class InstantiatedModel:
    def __init__(
        self,
        abstract_model: AbstractModel,
        regional_param_store: RegionalParamStore,
        data_store: DataStore,
        create_concrete_model: bool = True,
    ):
        self.abstract_model = abstract_model
        self.regional_param_store = regional_param_store
        self.data_store = data_store

        # Get the non-regional params (from config.yaml) from the RegionalParamStore
        self.params = regional_param_store.params

        # First, set the data functions in the abstract model
        self.set_data_functions()

        self.instance_data = self.get_param_values()

        if create_concrete_model:
            self.concrete_model = self.create_instance()

    def set_data_functions(self):
        # The data functions need to be changed in the abstract model
        # before initialization.
        self.abstract_model.baseline_emissions = self.data_store.data_object("baseline")
        self.abstract_model.population = (
            self.abstract_model.L
        ) = self.data_store.data_object("population")
        self.abstract_model.GDP = self.data_store.data_object("GDP")
        self.abstract_model.carbon_intensity = self.data_store.data_object(
            "carbon_intensity"
        )
        self.abstract_model.TFP = self.data_store.data_object("TFP")

    def create_instance(self):
        return self.abstract_model.create_instance(self.instance_data)

    def get_param_values(self):
        damage_module = self.params["model"]["damage module"]
        emissiontrade_module = self.params["model"]["emissiontrade module"]

        instance_data = {None: {}}

        # Main instance data
        self._set_instance_data_main(instance_data)

        ## Emission trade module:

        # Instance data for globalcostpool emission trade module:
        if emissiontrade_module == "globalcostpool":
            self._set_instance_data_emissiontrade_globalcostpool(instance_data)

        ## Damage module:

        # Instance data for RICE2010 damage/adaptation:
        if damage_module == "RICE2010":
            self._set_instance_data_rice2010(instance_data)

        # Instance data for RICE2012 damage/adaptation:
        if damage_module == "RICE2012":
            self._set_instance_data_rice2012(instance_data)

        # Instance data for COACCH damages:
        if damage_module == "COACCH":
            self._set_instance_data_coacch(instance_data)

        # For WITCH damage/adaption:
        # if damage_module == "WITCH":
        #     instance_data[None].update(self._instance_data_witch())

        # Raise warning for partially irreversible damages when damage module != RICE2010:
        if (
            damage_module == "RICE2010"
            and self.params["economics"]["damages"]["percentage reversible"] != 1
        ):
            raise NotImplementedError(
                'Partially irreversible damages not implemented for damage module "{}"'.format(
                    damage_module
                )
            )

        return instance_data

    ########################
    #
    # Private functions
    #
    ########################

    def _set_instance_data_main(self, instance_data) -> None:
        params = self.params
        t_start = params["time"]["start"]
        t_end = params["time"]["end"]
        dt = params["time"]["dt"]
        num_years = int(np.ceil((t_end - t_start) / dt)) + 1
        self.abstract_model.year = lambda t: t_start + t * dt
        year2100 = int((2100 - t_start) / dt)

        parameter_mapping = {
            "beginyear": V(t_start),
            "dt": V(dt),
            "tf": V(num_years - 1),
            "t": V(range(num_years)),
            "year2100": V(year2100),
            "regions": V(params["regions"].keys()),
            "baseline_carbon_intensity": V(
                params["emissions"]["baseline carbon intensity"]
            ),
            "budget": V(quant(params["emissions"]["carbonbudget"], "emissions_unit")),
            "inertia_regional": V(params["emissions"]["inertia"]["regional"]),
            "inertia_global": V(params["emissions"]["inertia"]["global"]),
            "global_min_level": V(
                quant(params["emissions"]["global min level"], "emissionsrate_unit")
            ),
            "regional_min_level": V(
                quant(params["emissions"]["regional min level"], "emissionsrate_unit")
            ),
            "no_pos_emissions_after_budget_year": V(
                params["emissions"]["not positive after budget year"]
            ),
            "non_increasing_emissions_after_2100": V(
                params["emissions"]["non increasing emissions after 2100"]
            ),
            "cumulative_emissions_trapz": V(
                params["emissions"]["cumulative_emissions_trapz"]
            ),
            "burden_sharing_regime": V(params["burden sharing"]["regime"]),
            "percapconv_year": V(params["burden sharing"]["percapconv_year"]),
            "T0": V(quant(params["temperature"]["initial"], "temperature_unit")),
            "temperature_target": V(
                quant(params["temperature"]["target"], "temperature_unit")
            ),
            "TCRE": V(
                quant(
                    params["temperature"]["TCRE"],
                    "(temperature_unit)/(emissions_unit)",
                )
            ),
            "LBD_rate": V(params["economics"]["MAC"]["rho"]),
            "LBD_scaling": V(quant("40 GtCO2", "emissions_unit")),
            "LOT_rate": V(0),
            "damage_scale_factor": V(params["economics"]["damages"]["scale factor"]),
            "fixed_adaptation": V(params["economics"]["adaptation"]["fixed"]),
            "perc_reversible_damages": V(
                params["economics"]["damages"]["percentage reversible"]
            ),
            "ignore_damages": V(params["economics"]["damages"]["ignore damages"]),
            "MAC_gamma": V(
                quant(
                    params["economics"]["MAC"]["gamma"],
                    "currency_unit/emissionsrate_unit",
                )
            ),
            "MAC_beta": V(params["economics"]["MAC"]["beta"]),
            "MAC_scaling_factor": self.regional_param_store.get(
                "MAC", params["economics"]["MAC"]["regional calibration factor"]
            ),
            "rel_abatement_costs_min_level": V(
                params["economics"]["MAC"]["rel_abatement_costs_min_level"]
            ),
            "init_capitalstock_factor": self.regional_param_store.get(
                "economics", "init_capital_factor"
            ),
            "alpha": V(params["economics"]["GDP"]["alpha"]),
            "dk": V(params["economics"]["GDP"]["depreciation of capital"]),
            "sr": V(params["economics"]["GDP"]["savings rate"]),
            "elasmu": V(params["economics"]["elasmu"]),
            "inequal_aversion": V(params["economics"]["inequal_aversion"]),
            "PRTP": V(params["economics"]["PRTP"]),
        }

        if "custom_mapping" in params["simulation"]:
            parameter_mapping.update(params["simulation"]["custom_mapping"])

        instance_data[None].update(parameter_mapping)

        # Set sea level rise parameters
        self._set_instance_data_slr(instance_data)

    def _set_instance_data_slr(self, instance_data) -> None:
        parameter_mapping = {
            # Sea level rise:
            "slr_thermal_equil": V(0.5),  # Equilibrium
            "slr_thermal_init": V(0.0920666936642),  # Initial condition
            "slr_thermal_adjust_rate": V(0.024076141150722),  # Adjustment rate
            "slr_gsic_melt_rate": V(0.0008),  # Melt rate
            "slr_gsic_total_ice": V(0.26),  # Total ice
            "slr_gsic_equil_temp": V(-1),  # Equil temp
            "slr_gis_melt_rate_above_thresh": V(1.11860082),  # Melt rate above threshol
            "slr_gis_init_melt_rate": V(0.6),  # Initial melt rate
            "slr_gis_init_ice_vol": V(7.3),  # Initial ice volume
        }

        instance_data[None].update(parameter_mapping)

    def _set_instance_data_emissiontrade_globalcostpool(self, instance_data) -> None:
        params = self.params
        parameter_mapping = {
            "min_rel_payment_level": V(
                params["economics"]["emission trade"]["min rel payment level"]
            ),
            "max_rel_payment_level": V(
                params["economics"]["emission trade"]["max rel payment level"]
            ),
        }

        instance_data[None].update(parameter_mapping)

    def _set_instance_data_rice2010(self, instance_data) -> None:
        params = self.params
        parameter_mapping = {
            "adapt_curr_level": V(params["economics"]["adaptation"]["curr_level"]),
            "damage_a1": self.regional_param_store.get("ADRICE2010", "a1"),
            "damage_a2": self.regional_param_store.get("ADRICE2010", "a2"),
            "damage_a3": self.regional_param_store.get("ADRICE2010", "a3"),
            "adapt_g1": self.regional_param_store.get("ADRICE2010", "g1"),
            "adapt_g2": self.regional_param_store.get("ADRICE2010", "g2"),
        }

        instance_data[None].update(parameter_mapping)

    def _set_instance_data_rice2012(self, instance_data) -> None:
        parameter_mapping = {
            "damage_a1": self.regional_param_store.get("ADRICE2012", "a1"),
            "damage_a2": self.regional_param_store.get("ADRICE2012", "a2"),
            "damage_a3": self.regional_param_store.get("ADRICE2012", "a3"),
            "adap1": self.regional_param_store.get("ADRICE2012", "nu1"),
            "adap2": self.regional_param_store.get("ADRICE2012", "nu2"),
            "adap3": self.regional_param_store.get("ADRICE2012", "nu3"),
            "adapt_rho": V(0.5),
            "SLRdam1": self.regional_param_store.get("ADRICE2012", "slrdam1"),
            "SLRdam2": self.regional_param_store.get("ADRICE2012", "slrdam2"),
        }

        instance_data[None].update(parameter_mapping)

    def _set_instance_data_coacch(self, instance_data) -> None:
        try:
            damage_quantile = self.params["economics"]["damages"]["quantile"]
        except KeyError:
            damage_quantile = 0.5

        combined_slr_nonslr_damages = self.params["economics"]["damages"][
            "coacch_combined_slr_nonslr_damages"
        ]
        slr_withadapt = self.params["economics"]["damages"]["coacch_slr_withadapt"]
        adapt_prfx = "Ad" if slr_withadapt else "NoAd"
        prfx = f"SLR-{adapt_prfx}"

        V_region = lambda x: {region: x for region in self.params["regions"]}

        if combined_slr_nonslr_damages:
            parameter_mapping = {
                # Combined non-SLR and SLR damages are always quadratic
                "damage_noslr_form": V_region("Robust-Quadratic"),
                "damage_noslr_b1": self.regional_param_store.get(
                    "COACCH", f"combined_b1_{adapt_prfx}-q{damage_quantile}"
                ),
                "damage_noslr_b2": self.regional_param_store.get(
                    "COACCH", f"combined_b2_{adapt_prfx}-q{damage_quantile}"
                ),
                "damage_noslr_b3": V_region(0),
                "damage_noslr_a": V_region(1),
                # SLR damages (zero but need to be defined)
                "damage_slr_form": V_region("Robust-Linear"),
                "damage_slr_b1": V_region(0),
                "damage_slr_b2": V_region(0),
                "damage_slr_b3": V_region(0),
                "damage_slr_a": V_region(0),
            }

        else:
            factor_noslr = f"NoSLR_a (q={damage_quantile})"
            factor_slr_ad = f"{prfx}_a (q={damage_quantile})"

            parameter_mapping = {
                # Non-SLR damages:
                "damage_noslr_form": self.regional_param_store.get(
                    "COACCH", "NoSLR_form"
                ),
                "damage_noslr_b1": self.regional_param_store.get("COACCH", "NoSLR_b1"),
                "damage_noslr_b2": self.regional_param_store.get("COACCH", "NoSLR_b2"),
                "damage_noslr_b3": self.regional_param_store.get("COACCH", "NoSLR_b3"),
                "damage_noslr_a": self.regional_param_store.get("COACCH", factor_noslr),
                # SLR damages:
                "damage_slr_form": self.regional_param_store.get(
                    "COACCH", f"{prfx}_form"
                ),
                "damage_slr_b1": self.regional_param_store.get("COACCH", f"{prfx}_b1"),
                "damage_slr_b2": self.regional_param_store.get("COACCH", f"{prfx}_b2"),
                "damage_slr_b3": self.regional_param_store.get("COACCH", f"{prfx}_b3"),
                "damage_slr_a": self.regional_param_store.get("COACCH", factor_slr_ad),
            }

        instance_data[None].update(parameter_mapping)

    # def _instance_data_witch(self) -> dict:
    #     return {
    #         "damage_omega1_pos": self.data_store.get_regional("damages", "omega1_pos"),
    #         "damage_omega1_neg": self.data_store.get_regional("damages", "omega1_neg"),
    #         "damage_omega2_pos": self.data_store.get_regional("damages", "omega2_pos"),
    #         "damage_omega2_neg": self.data_store.get_regional("damages", "omega2_neg"),
    #         "damage_omega3_pos": self.data_store.get_regional("damages", "omega3_pos"),
    #         "damage_omega3_neg": self.data_store.get_regional("damages", "omega3_neg"),
    #         "damage_omega4_pos": self.data_store.get_regional("damages", "omega4_pos"),
    #         "damage_omega4_neg": self.data_store.get_regional("damages", "omega4_neg"),
    #         "adapt_omega_eff_ada": self.data_store.get_regional(
    #             "adaptation", "omega_eff_ada"
    #         ),
    #         "adapt_omega_act": self.data_store.get_regional("adaptation", "omega_act"),
    #         "adapt_omega_eff_act": self.data_store.get_regional(
    #             "adaptation", "omega_eff_act"
    #         ),
    #         "adapt_omega_rada": self.data_store.get_regional(
    #             "adaptation", "omega_rada"
    #         ),
    #         "adapt_rho_ada": self.data_store.get_regional("adaptation", "rho_ada"),
    #         "adapt_rho_act": self.data_store.get_regional("adaptation", "rho_act"),
    #         "adapt_eps": self.data_store.get_regional("adaptation", "eps"),
    #     }



"""Define region mappers, used to transform one regional definition to another.

Currently, only `mean` is used. In future versions, this will be extended
and automatic region mapping will be used by transforming to country level first."""

import os
import typing
import pandas as pd


class RegionMapper:
    """
    Maps a region type to a different region type
    (RICE to IMAGE regions for example)
    """

    def __init__(self, regiontype1, regiontype2, mapping_file):
        full_filename = os.path.join(
            os.path.dirname(__file__),
            "../../inputdata/regions",
            mapping_file,
        )
        # Read region mapping
        self.mapping = pd.read_csv(full_filename)

        # Check if both region types are present as columns
        if (
            regiontype1 not in self.mapping.columns
            or regiontype2 not in self.mapping.columns
        ):
            raise KeyError(
                "Region mappers: file {} doesn't contain columns {} and {}.".format(
                    mapping_file, regiontype1, regiontype2
                )
            )

    def map_regions(
        self, data: pd.DataFrame, regiontype_from, regiontype_to, region_column="region"
    ):
        # Merge data with mapping region
        data_merged = data.merge(
            self.mapping[[regiontype_from, regiontype_to]],
            left_on=region_column,
            right_on=regiontype_from,
            how="outer",
        ).drop(columns=["region", regiontype_from])

        # Group by region_to (if there are multiple regions mapping to one,
        # take average, or the first value for non-numeric columns)
        aggregation_methods = {
            col_name: "mean" if pd.api.types.is_numeric_dtype(column) else "first"
            for col_name, column in data_merged.items()
            if col_name != regiontype_to
        }
        data_grouped = (
            data_merged.groupby(regiontype_to)
            .agg(aggregation_methods)
            .reset_index()
            .rename(columns={regiontype_to: "region"})
        )

        return data_grouped


REGION_MAPPERS: typing.Dict[tuple, RegionMapper] = {}


def register_region_mappers(regiontype1, regiontype2, mapping_file):
    mapper = RegionMapper(regiontype1, regiontype2, mapping_file)

    # Save for both regiontype1 to regiontype2 and vice versa
    REGION_MAPPERS[(regiontype1, regiontype2)] = mapper
    REGION_MAPPERS[(regiontype2, regiontype1)] = mapper


register_region_mappers("IMAGE26", "ADRICE2010", "IMAGE26_ADRICE2010.csv")
register_region_mappers("IMAGE26", "ADRICE2012", "IMAGE26_ADRICE2012.csv")
register_region_mappers("IMAGE26", "COACCH", "IMAGE26_COACCH.csv")



"""
Utils to calculate TFP using the Cobb-Douglas equation
"""

import numpy as np
from mimosa.common.data.utils import UnitValues
from mimosa.common.regional_params import RegionalParamStore


def get_TFP(
    region,
    years,
    gdp_all_regions,
    population_all_regions,
    regional_param_store: RegionalParamStore,
) -> UnitValues:
    params = regional_param_store.params
    # quant = data_store.quant
    tfp = []
    dt = years[1] - years[0]

    # Parameters
    alpha = params["economics"]["GDP"]["alpha"]
    depr_cap = params["economics"]["GDP"]["depreciation of capital"]
    savings_rate = params["economics"]["GDP"]["savings rate"]

    # Get data
    gdp_data = gdp_all_regions[region]
    population_data = population_all_regions[region]

    # Initialise capital
    capital = (
        regional_param_store.getregional("economics", "init_capital_factor", region)
        * gdp_data.yvalues[0]
    )

    for gdp, pop in zip(gdp_data.yvalues, population_data.yvalues):
        tfp.append(gdp / calc_GDP(1, pop, capital, alpha))
        dKdt = calc_dKdt(capital, depr_cap, savings_rate * gdp, dt)
        capital = dKdt * dt + capital

    return UnitValues(gdp_data.xvalues, np.array(tfp))


def calc_dKdt(K, dk, I, dt):
    return ((1 - dk) ** dt - 1) / dt * K + I


def calc_GDP(TFP, L, K, alpha):
    return TFP * L ** (1 - alpha) * K**alpha



from setuptools import setup

with open("README.md") as fh:
    description = fh.read()

setup(
    name="mimosa",
    version="0.1.0",
    author="Kaj-Ivar van der Wijst",
    author_email="k.i.vanderwijst@gmail.com",
    packages=["mimosa"],
    # url="http://pypi.python.org/pypi/PackageName/",
    license="LICENSE.txt",
    description="MIMOSA: Integrated Assessment Model for Cost-Benefit Analysis",
    long_description=description,
    long_description_content_type="text/markdown",
    install_requires=["numpy", "pandas", "pyomo", "pint", "pyyaml"],
    include_package_data=True,
)



"""
Uses the Pint package to parse quantities with units.
The custom units are loaded from a text file.
"""
import os
from pyomo.environ import units as pyomo_units

from .utils import load_yaml


pyomo_units.load_definitions_from_file(
    os.path.join(os.path.dirname(__file__), "../inputdata/config", "extra_units.txt")
)

DEFAULT_UNITS = load_yaml("default_units.yaml")


class Quantity:
    """
    The Quantity object creates a callable object which parses quantities
    with units (see __call__ usage).

    Args:
        params (dict): contains all Param values

    """

    def __init__(self):
        self.default_units = DEFAULT_UNITS
        self._pint_registry = pyomo_units.pint_registry
        self._pyomo_units_container = pyomo_units

    def __call__(self, *args, only_magnitude=True, can_be_false=True):
        """Usage:
        - Quant(string, to_unit):        Quant('20 GtCO2/yr', 'emissionsrate_unit')
        - Quant(value, unit, to_unit):   Quant(20, 'GtCO2/yr', 'emissionsrate_unit')
        """

        if len(args) not in [2, 3]:  # If number of arguments is not 2 or 3
            raise RuntimeError("Incorrect number of arguments to Q")

        if args[0] is False and can_be_false:
            # Sometimes a quantity (like carbon budget) can be skipped by setting to False
            return False

        # First parse target units with default units
        target_unit = self._parse_default_units(self._custom_replace(args[-1]))

        if len(args) == 2:
            string = self._custom_replace(str(args[0]))
            quantity = self._pint_registry.Quantity(string)
        elif len(args) == 3:
            value = args[0]
            unit = self._custom_replace(str(args[1]))
            quantity = self._pint_registry.Quantity(value, unit)
        else:
            raise Exception("Wrong usage of Quant function")

        if only_magnitude:
            return quantity.to(target_unit).magnitude
        return quantity.to(target_unit)

    def __repr__(self):
        units = ", ".join(
            [f"{key}: {value}" for key, value in self.default_units.items()]
        )
        return f"Quantity with units: {units}"

    def unit(self, unit_str: str, pyomo=True):
        """Returns a Pyomo unit object given a string with units.

        If pyomo=false, returns a Pint unit object.

        First it replaces custom default units:
            - currency_unit
            - emissionsrate_unit
            - emissions_unit
            - temperature_unit
            - population_unit
        to the respective default unit as defined in default_units.yaml
        (which is also stored in this object as `self.default_units`)
        """

        default_units_replaced = self._parse_default_units(unit_str)
        if pyomo:
            return getattr(self._pyomo_units_container, default_units_replaced)

        return self._pint_registry.Unit(default_units_replaced)

    ####### Private functions #######

    def _parse_default_units(self, units):
        for key, value in self.default_units.items():
            # Add brackets () to avoid order of operation problems with compound units
            units = units.replace(key, f"({value})")
        # Bugfix for a Pint bug where aliases are treated wrongly
        units = units.replace("year", "yr")
        return units

    @staticmethod
    def _custom_replace(units):
        return units.replace("US$", "USD").replace("t CO2", "tCO2")


quant = Quantity()



"""
Plot functions. Not necessary if dashboard is used.
"""
import numpy as np
import pandas as pd
import plotly.express as px
from plotly.subplots import make_subplots
from pyomo.environ import value

COLORS_PBL = [
    "#00AEEF",
    "#808D1D",
    "#B6036C",
    "#FAAD1E",
    "#3F1464",
    "#7CCFF2",
    "#F198C1",
    "#42B649",
    "#EE2A23",
    "#004019",
    "#F47321",
    "#511607",
    "#BA8912",
    "#78CBBF",
    "#FFF229",
    "#0071BB",
]


class Plot:
    """
    Wrapper for a Plotly figure.

    First it creates a subplot figure, then using the function `add`,
    a trace can be added given a Pyomo variable.
    """

    def __init__(self, m, numrows=4, globalrows=None, coltitles=None, **kwargs):
        if globalrows is None:
            globalrows = [4]
        self.m = m
        self.regions = m.regions
        self.years = m.year(np.array(m.t))
        self.t = m.t

        n = len(m.regions)
        specs = [[{"secondary_y": True}] * n for _ in range(numrows)]
        for row in globalrows:
            specs[row - 1] = [{"colspan": n, "secondary_y": True}] + [None] * (n - 1)

        self.fig = make_subplots(
            numrows,
            n,
            shared_yaxes=True,
            shared_xaxes="rows",
            specs=specs,
            subplot_titles=coltitles if coltitles is not None else list(m.regions),
            **kwargs
        )

        # Make sure all secondary yaxes are matched per row
        for i in range(numrows):
            ref = next(self.fig.select_yaxes(row=i + 1, col=1, secondary_y=True))
            ref_name = ref.plotly_name.replace("axis", "")
            for axis in self.fig.select_yaxes(row=i + 1, secondary_y=True):
                axis.matches = ref_name

        self.curr_values = {}

    def _value(self, var, region, is_fct):
        if is_fct:
            return [var(t, region) for t in self.years]
        else:
            return [value(var[t, region]) for t in self.t]

    def _value_global(self, var, is_fct):
        if is_fct:
            return [var(t) for t in self.t]
        else:
            return [value(var[t]) for t in self.t]

    def add_scatter(self, values, name, color, row, col, showlegend, **kwargs):
        self.fig.add_scatter(
            x=self.years,
            y=values,
            name=name,
            legendgroup=name,
            line_color=color,
            row=row,
            col=col,
            showlegend=showlegend,
            **(
                {"line_dash": "dot" if kwargs.get("secondary_y", False) else "solid"}
                if "line_dash" not in kwargs
                else {}
            ),
            **kwargs
        )

    def add(self, var, is_regional=True, is_fct=False, name=None, row=1, **kwargs):

        if not is_fct and name is None:
            name = var.name
        new = name not in self.curr_values

        if new:
            self.curr_values[name] = len(self.curr_values)
        color = COLORS_PBL[self.curr_values[name] % len(COLORS_PBL)]

        if is_regional:
            for i, r in enumerate(self.regions):
                self.add_scatter(
                    self._value(var, r, is_fct),
                    name,
                    color,
                    row,
                    col=i + 1,
                    showlegend=new and (i == 0),
                    **kwargs
                )
        else:
            self.add_scatter(
                self._value_global(var, is_fct),
                name,
                color,
                row,
                col=1,
                showlegend=new,
                **kwargs
            )

    def set_layout(self):
        self.fig.update_xaxes(range=[self.years[0], 2100])
        for i in range(1, len(self.regions)):
            self.fig.update_yaxes(secondary_y=True, showticklabels=False, col=i)
        self.fig.update_traces(mode="lines")
        self.fig.update_layout(
            margin={"t": 50, "l": 50, "r": 50, "b": 50},
            legend={"yanchor": "middle", "y": 0.5},
        )

    def set_yaxes_titles(self, titles):
        for i, title in enumerate(titles):
            self.fig.update_yaxes(title=title, row=i + 1, col=1, secondary_y=False)

    def show(self):
        self.set_layout()
        return self.fig


def visualise_ipopt_output(output_file):
    file = open(output_file, "r")
    in_iterations = False

    iterations = []

    for line in file:
        if line.startswith("iter "):
            in_iterations = True
        if line.strip() == "":
            in_iterations = False

        if in_iterations and line.startswith(" "):
            split = line.strip().split(" ")[:4]

            # Check if it was a recovered iteration
            if "r" in split[0]:
                recovered = True
                split = split[0].split("r") + split[1:-1]
            else:
                recovered = False
            iterations.append(split + [recovered])

    iterations_df = pd.DataFrame(
        iterations, columns=["iter", "objective", "inf_pr", "inf_du", "recovered"]
    )
    iterations_df = iterations_df.astype(
        {"iter": int, "objective": float, "inf_pr": float, "inf_du": float}
    )

    file.close()

    fig = (
        px.scatter(
            iterations_df,
            x="iter",
            y=["objective", "inf_pr", "inf_du"],
            facet_row="variable",
            color="recovered",
        )
        .update_yaxes(matches=None, type="log")
        .update_yaxes(row=3, type="linear")
    )

    fig.write_html(output_file.split(".")[0] + ".html", include_plotlyjs="cdn")
